(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2010-2015)
Author:  Julian Nagele <julian.nagele@uibk.ac.at> (2012-2015)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2010-2015)
Author:  Sarah Winkler <sarah.winkler@uibk.ac.at> (2013, 2014)
Author:  Thomas Sternagel <thomas.sternagel@uibk.ac.at> (2012)
License: LGPL (see file COPYING.LESSER)
*)
theory CPF_Parser
imports
  Proof_Checker
  "$AFP/XML/Xmlt"
begin

definition var :: "('a, string) term xmlt" where "var x = (Xmlt.text ''var'' x \<guillemotright>= (return \<circ> Var))"
hide_const (open) var

definition plain_var :: "string xmlt" where "plain_var x = (Xmlt.text ''var'' x)"
hide_const (open) plain_var  

partial_function (sum_bot)
  "term" :: "'a xmlt \<Rightarrow> ('a, string) term xmlt"
where
  [code]: "term xml2name x = Xmlt.options [
    (''var'', CPF_Parser.var),
    (''funapp'', Xmlt.many1 ''funapp'' xml2name (
      Xmlt.singleton ''arg'' (term xml2name) id) Fun)
  ] x"  
hide_const (open) "term"

partial_function (sum_bot) ctxt :: "'a xmlt \<Rightarrow> ('a, string) ctxt xmlt" where
  [code]: "ctxt xml2name x = Xmlt.options [
     (''box'', Xmlt.leaf ''box'' Hole),  
     (''funContext'', Xmlt.tuple4 ''funContext'' 
       xml2name 
       (Xmlt.many ''before'' (CPF_Parser.term xml2name) id)
       (ctxt xml2name)
       (Xmlt.many ''after'' (CPF_Parser.term xml2name) id)
       More)
   ] x"  

definition flat_contexts :: "'a xmlt \<Rightarrow> ('a, string) ctxt list xmlt" where
  "flat_contexts xml2name = Xmlt.many ''flatContexts'' (CPF_Parser.ctxt xml2name) id"
hide_const (open) flat_contexts

definition subst :: "'a xmlt \<Rightarrow> ('a, string) substL xmlt" where
  "subst xml2name = Xmlt.many ''substitution''
    (Xmlt.pair ''substEntry'' 
      (Xmlt.text ''var'')
      (CPF_Parser.term xml2name)
      Pair
    ) id"
hide_const (open) subst

primrec list2position :: "nat list \<Rightarrow> pos" where
  "list2position [] = \<epsilon>"
| "list2position (n # ns) = n <# list2position ns"

definition position :: "nat xmlt" where
  "position \<equiv> Xmlt.change (Xmlt.nat ''position'') (\<lambda>n. n - 1)"
hide_const (open) position

definition pos :: "pos xmlt" where
  "pos \<equiv> Xmlt.many ''positionInTerm'' CPF_Parser.position list2position"
hide_const (open) pos


definition afs :: "'a xmlt \<Rightarrow> 'a afs_list xmlt" where
  "afs xml2name \<equiv> Xmlt.many ''argumentFilter'' (Xmlt.triple ''argumentFilterEntry''
    xml2name
    (Xmlt.nat ''arity'')
    (Xmlt.options [
      (''collapsing'', Xmlt.change (Xmlt.nat ''collapsing'') (\<lambda>i. Collapse (i - 1))),
      (''nonCollapsing'', Xmlt.many ''nonCollapsing'' CPF_Parser.position AFList)
    ])
    (\<lambda>f a e. ((f, a), e))
  ) id"
hide_const (open) afs

definition proj :: "'a xmlt \<Rightarrow> 'a projL xmlt" where
  "proj xml2name \<equiv> Xmlt.change (CPF_Parser.afs xml2name) (\<lambda>afl. Projection 
    (map (\<lambda>(fa, e). case e of 
      Collapse i \<Rightarrow> Pair fa i
    | _ \<Rightarrow> Pair fa (snd fa)) afl))" 

definition rule :: "'a xmlt \<Rightarrow> ('a, string) rule xmlt" where
  "rule xml2name = Xmlt.pair ''rule''
    (Xmlt.singleton ''lhs'' (CPF_Parser.term xml2name) id)
    (Xmlt.singleton ''rhs'' (CPF_Parser.term xml2name) id)
    Pair"
hide_const (open) rule 

definition
  relstep :: "'a xmlt \<Rightarrow> (pos \<times> ('a, string) rule \<times> bool \<times> ('a, string) term) xmlt"
where
  "relstep xml2name = xml3to4elements ''rewriteStep''
    CPF_Parser.pos
    (CPF_Parser.rule xml2name)
    (Xmlt.leaf ''relative'' ())
    (CPF_Parser.term xml2name)
    (\<lambda>p r rel t. (p, r, rel = None, t))"
hide_const (open) relstep

definition
  rstep :: "'a xmlt \<Rightarrow> (pos \<times> ('a, string) rule \<times> ('a, string) term) xmlt"
where
  "rstep xml2name = Xmlt.triple ''rewriteStep''
    CPF_Parser.pos
    (CPF_Parser.rule xml2name)
    (CPF_Parser.term xml2name)
    (\<lambda>p r t. (p, r, t))"
hide_const (open) rstep

definition
  estep :: "'a xmlt \<Rightarrow> (pos \<times> ('a, string) rule \<times> bool \<times> ('a, string) term) xmlt"
where
  "estep xml2name = Xmlt.tuple4 ''equationStep''
    CPF_Parser.pos
    (CPF_Parser.rule xml2name)
    (Xmlt.options [
       (''leftRight'', Xmlt.leaf ''leftRight'' True),
       (''rightLeft'', Xmlt.leaf ''rightLeft'' False)
    ])
    (CPF_Parser.term xml2name)
    (\<lambda>p r b t. (p, r, b, t))"
hide_const (open) estep

definition
  relsteps :: "'a xmlt \<Rightarrow> (('a,string)term \<times> ('a,string)prseq) xmlt"
where
  "relsteps xml2name \<equiv> Xmlt.many1 ''rewriteSequence'' 
    (Xmlt.singleton ''startTerm'' (CPF_Parser.term xml2name) id)
    (CPF_Parser.relstep xml2name)
    Pair"
hide_const (open) relsteps

definition
  rsteps :: "'a xmlt \<Rightarrow> (('a,string)term \<times> ('a,string)rseq) xmlt"
where
  "rsteps xml2name \<equiv> Xmlt.many1 ''rewriteSequence'' 
    (Xmlt.singleton ''startTerm'' (CPF_Parser.term xml2name) id)
    (CPF_Parser.rstep xml2name)
    Pair"
hide_const (open) rsteps


definition pat_eqv_prf :: "'a xmlt \<Rightarrow> ('a,string)pat_eqv_prf xmlt"
  where "pat_eqv_prf xml2name \<equiv> let sub = CPF_Parser.subst xml2name in
   Xmlt.singleton ''patternEquivalence''
    (Xmlt.options [
       (''domainRenaming'', Xmlt.singleton ''domainRenaming'' sub Pat_Dom_Renaming),
       (''irrelevant'', Xmlt.pair ''irrelevant''  sub sub Pat_Irrelevant),
       (''simplification'', Xmlt.pair ''simplification'' sub sub Pat_Simplify)
     ]) id"
hide_const (open) pat_eqv_prf

definition pat_term :: "'f xmlt \<Rightarrow> (('f,string)term \<times> ('f,string)substL \<times> ('f,string)substL) xmlt"
  where "pat_term xml2name \<equiv> Xmlt.triple ''patternTerm''
            (CPF_Parser.term xml2name)
            (CPF_Parser.subst xml2name)
            (CPF_Parser.subst xml2name)
            (\<lambda> t s1 s2. (t,s1,s2))"


abbreviation (input) pat_rule_fun where "pat_rule_fun xml2name x pat \<equiv> (let 
   sub = CPF_Parser.subst xml2name;
   pt = pat_term xml2name;
   var = Xmlt.text ''var''
 in 
   Xmlt.triple ''patternRule'' pt pt
    (Xmlt.options [
       (''originalRule'', Xmlt.pair ''originalRule'' (CPF_Parser.rule xml2name) (Xmlt.bool ''isPair'')  Pat_OrigRule),
       (''initialPumping'', Xmlt.triple ''initialPumping'' pat sub sub Pat_InitPump),
       (''initialPumpingContext'', Xmlt.tuple4 ''initialPumpingContext'' pat sub CPF_Parser.pos var Pat_InitPumpCtxt),
       (''equivalence'', Xmlt.triple ''equivalence'' pat (Xmlt.options [
                  (''left'',Xmlt.leaf ''left'' True),
                  (''right'',Xmlt.leaf ''right'' False)
             ]) (CPF_Parser.pat_eqv_prf xml2name) Pat_Equiv),
       (''narrowing'', Xmlt.triple ''narrowing'' pat pat CPF_Parser.pos Pat_Narrow),
       (''instantiation'', Xmlt.triple ''instantiation'' pat sub (Xmlt.options [
                  (''base'',Xmlt.leaf ''base'' Pat_Base),
                  (''pumping'',Xmlt.leaf ''pumping'' Pat_Pump),
                  (''closing'',Xmlt.leaf ''closing'' Pat_Close)
             ]) Pat_Inst),
       (''instantiationPumping'', Xmlt.pair ''instantiationPumping'' pat 
             (Xmlt.nat ''power'')
             Pat_Exp_Sigma),
       (''rewriting'', Xmlt.triple ''rewriting'' pat (CPF_Parser.rsteps xml2name)  (Xmlt.options [
                  (''base'',Xmlt.leaf ''base'' (Pat_Base,[])),
                  (''pumping'',Xmlt.singleton ''pumping'' var (Pair Pat_Pump)),
                  (''closing'',Xmlt.singleton ''closing'' var (Pair Pat_Close))
             ]) (\<lambda> p seq (po,va). Pat_Rewr p seq po va))
     ]) (\<lambda> _ _ z. z) x)" 

partial_function (sum_bot) pat_rule_prf :: "'a xmlt \<Rightarrow> ('a,string)pat_rule_prf xmlt"
  where "pat_rule_prf xml2name x = pat_rule_fun xml2name x (pat_rule_prf xml2name)"

(* one cannot put the 'let pat = ... ' directly into partial function, since then the monotonicity fails *)
lemma pat_rule_prf [code]: 
  "pat_rule_prf xml2name x = (let pat = pat_rule_prf xml2name in pat_rule_fun xml2name x pat)"
     unfolding pat_rule_prf.simps[of _ x] Let_def by simp

hide_const (open) pat_rule_prf


definition
  conversion :: "'a xmlt \<Rightarrow> (('a,string)eseq) xmlt"
where
  "conversion xml2name \<equiv> Xmlt.many1 ''conversion'' 
    (Xmlt.singleton ''startTerm'' (CPF_Parser.term xml2name) id)
    (CPF_Parser.estep xml2name)
    (\<lambda> _ x. x)"
hide_const (open) conversion

definition subsumption_proof :: "'a xmlt \<Rightarrow> ('a,string)subsumption_proof xmlt"
  where "subsumption_proof xml2name \<equiv> Xmlt.many ''subsumptionProof''
     (Xmlt.pair ''ruleSubsumptionProof''
        (CPF_Parser.rule xml2name)
        (CPF_Parser.conversion xml2name)
        Pair
    ) id"
hide_const (open) subsumption_proof

definition joinable_critical_pairs :: "'a xmlt \<Rightarrow> ('a,string)crit_pair_joins xmlt"
  where "joinable_critical_pairs xml2name \<equiv> let rew = CPF_Parser.rsteps xml2name in
    Xmlt.many ''joinableCriticalPairs'' 
      (Xmlt.pair ''joinableCriticalPair'' rew rew (\<lambda> (s,sseq) (t,tseq). (s,sseq,t,tseq))) id"

definition wcr_proof :: "'a xmlt \<Rightarrow> 'a join_info xmlt"
  where "wcr_proof xml2name \<equiv> Xmlt.singleton ''wcrProof'' 
    (Xmlt.options [
       (''joinableCriticalPairs'', Xmlt.change (joinable_critical_pairs xml2name) Guided),
       (''joinableCriticalPairsBFS'', Xmlt.change (Xmlt.nat ''joinableCriticalPairsBFS'') Join_BFS),
       (''joinableCriticalPairsAuto'', Xmlt.leaf ''joinableCriticalPairsAuto'' Join_NF)
  ]) id"

definition
  forbidden_pattern :: "'a xmlt \<Rightarrow> ((('a, string) ctxt \<times> ('a, string) term \<times> Forbidden_Patterns.location) xmlt)"
where
  "forbidden_pattern xml2name xml \<equiv> ((Xmlt.triple ''forbiddenPattern''
    (CPF_Parser.term xml2name)
    CPF_Parser.pos
    (Xmlt.options [
       (''here'', Xmlt.leaf ''here'' Forbidden_Patterns.H),
       (''above'', Xmlt.leaf ''above'' Forbidden_Patterns.A),
       (''below'', Xmlt.leaf ''below'' Forbidden_Patterns.B),
       (''right'', Xmlt.leaf ''below'' Forbidden_Patterns.R)
    ]) 
    (\<lambda> t p l. (t,p,l)) xml) \<guillemotright>= 
   (\<lambda> (t,p,l).
   if p \<in> poss t then return (ctxt_of_pos_term p t, t |_ p, l) else error (''position does not exist in term'')))"
hide_const (open) forbidden_pattern

definition
  forbidden_patterns :: "'a xmlt \<Rightarrow> ((('a, string) ctxt \<times> ('a, string) term \<times> Forbidden_Patterns.location) list) xmlt"
where
  "forbidden_patterns xml2name \<equiv> Xmlt.many ''forbiddenPatterns''
    (CPF_Parser.forbidden_pattern xml2name)
    id"
hide_const (open) forbidden_patterns

definition 
  loop ::
    "'a xmlt \<Rightarrow>
    (('a, string) term \<times> ('a, string) prseq \<times> ('a, string) substL \<times> ('a, string) ctxt) xmlt"
where
  "loop xml2name \<equiv> Xmlt.triple ''loop''
    (CPF_Parser.relsteps xml2name)
    (CPF_Parser.subst xml2name)
    (CPF_Parser.ctxt xml2name)
    (\<lambda>(s, rseq) \<sigma> c. (s, rseq, \<sigma>, c))"
hide_const (open) loop

definition
  start_term :: "'a xmlt \<Rightarrow> ('a, string) term \<Rightarrow> ('a, string) term xmlt"
where
  "start_term xml2name t = Xmlt.singleton ''startTerm'' (\<lambda>xml. (
     CPF_Parser.term xml2name xml \<guillemotright>= (\<lambda>s. if s = t
       then return t
       else error (''<startTerm> does not match lhs'')
     )
   )) id"
hide_const (open) start_term

definition
  rseq ::
    "'a xmlt \<Rightarrow> 'a proj \<Rightarrow> ('a, string) rule \<Rightarrow> (('a,string)rule \<times> ('a,string)rseq)xmlt"
where
  "rseq xml2name \<pi> r = Xmlt.many1 ''rewriteSequence''
     (CPF_Parser.start_term xml2name (proj_term \<pi> (fst r)))
     (CPF_Parser.rstep xml2name)
     (\<lambda>_ rseq. (r, rseq))"
hide_const (open) rseq

definition
  projected_rseq ::
    "'a xmlt \<Rightarrow> 'a proj \<Rightarrow>
    (('a, string) rule \<times> ('a,string)rseq) xmlt"
where
  "projected_rseq xml2name \<pi> xml = (let tag = ''projectedRewriteSequence'' in 
    case xml of
      XML name atts cs \<Rightarrow> 
        if name = tag \<and> atts = [] \<and> length cs = 2
        then (CPF_Parser.rule xml2name (cs ! 0) \<guillemotright>= (\<lambda>r. CPF_Parser.rseq xml2name \<pi> r (cs ! 1)))
        else Xmlt.fail tag xml
    | _ \<Rightarrow> Xmlt.fail tag xml)"
hide_const (open) projected_rseq

definition rules :: "'a xmlt \<Rightarrow> ('a, string) rules xmlt" where
  "rules xml2name = Xmlt.many ''rules'' (CPF_Parser.rule xml2name) id"
hide_const (open) rules


partial_function (sum_bot)
  arith_fun :: "arithFun xmlt"
where
  [code]: "arith_fun xml = Xmlt.singleton ''arithFunction'' (Xmlt.options [ 
     (''natural'', Xmlt.change (Xmlt.nat ''natural'') Const),
     (''variable'', Xmlt.change (Xmlt.nat ''variable'') (\<lambda>n::nat. Arg (n - 1))),
     (''sum'', Xmlt.many ''sum'' arith_fun Sum),
     (''product'', Xmlt.many ''product'' arith_fun Prod),
     (''min'', Xmlt.many ''min'' arith_fun Min),
     (''max'', Xmlt.many ''max'' arith_fun Semantic_Labeling_Carrier.Max),
     (''ifEqual'', Xmlt.tuple4 ''ifEqual'' arith_fun arith_fun arith_fun arith_fun IfEqual)
   ]) id  xml"

definition
  sl_variant :: "'f xmlt \<Rightarrow> ('f,string) sl_variant xmlt"
where
  "sl_variant xml2name \<equiv> Xmlt.singleton ''model'' (Xmlt.options [
     (''finiteModel'', xml1or2many_elements ''finiteModel''
         (Xmlt.change (Xmlt.nat ''carrierSize'') (\<lambda> n :: nat. n - 1))
         (Xmlt.singleton ''tupleOrder'' (Xmlt.leaf ''pointWise'' ()) id)
         (Xmlt.triple ''interpret''
           xml2name
           (Xmlt.nat ''arity'')
           CPF_Parser.arith_fun
           (\<lambda>f a inter. ((f, a), inter)))
         (\<lambda>n model xs. if model = None then Finitelab (SL_Inter n xs) else QuasiFinitelab (SL_Inter n xs) [])),
     (''rootLabeling'',Xmlt.guard (\<lambda> x. Xml.num_children x = 0) 
            (Xmlt.leaf ''rootLabeling'' (Rootlab None))
            (Xmlt.singleton ''rootLabeling'' xml2name (\<lambda> f. Rootlab (Some (f,1))))
        )
   ]) id"
hide_const (open) sl_variant

context
  fixes C :: "'a ring" (structure)
begin
fun sum_lpoly :: "'a list \<Rightarrow> 'a list \<Rightarrow> 'a list" where
  "sum_lpoly [] ys = ys"
| "sum_lpoly xs [] = xs"
| "sum_lpoly (x # xs) (y # ys) = (x \<oplus> y) # sum_lpoly xs ys"

fun
  lpoly_of :: 
    "(nat, 'a) tpoly \<Rightarrow> string +\<^sub>\<bottom> ('a \<times> 'a list)"
where
  "lpoly_of (PNum i) = return (i, [])"
| "lpoly_of (PVar x) = return (\<zero>, replicate x \<zero> @ [\<one>])"
| "lpoly_of (PSum []) = return (\<zero>, [])"
| "lpoly_of (PSum (p # ps)) = (
     lpoly_of  p 
  \<guillemotright>= (\<lambda>(cp, ncp). lpoly_of (PSum ps)
  \<guillemotright>= (\<lambda>(cq, ncq). return (cp \<oplus> cq, sum_lpoly ncp ncq))))"
| "lpoly_of (PMult []) = return (\<one>, [])"
| "lpoly_of (PMult (p # ps)) = (
     lpoly_of p
  \<guillemotright>= (\<lambda>(cp, ncp). lpoly_of (PMult ps)
  \<guillemotright>= (\<lambda>(cq, ncq). if (Ball (set ncp) (op = \<zero>))
       then return (cp \<otimes> cq, map (op \<otimes> cp) ncq)
       else (if (Ball (set ncq) (op = \<zero>))
         then return (cp \<otimes> cq, map (\<lambda> x. x \<otimes> cq) ncp)
         else error ''cannot transform non-linear polynomial to linear polynomial''))))"

primrec fit_length :: "nat \<Rightarrow> 'a list \<Rightarrow> 'a list" where
  "fit_length (Suc n) as = (case as of 
    [] \<Rightarrow> replicate (Suc n) \<zero>
  | b # bs \<Rightarrow> b # fit_length n bs)"
| "fit_length 0 _ = []"
end

partial_function (sum_bot) polynomial :: "'a xmlt \<Rightarrow> (nat, 'a) tpoly xmlt" where
  [code]: "polynomial xml2coeff x = Xmlt.singleton ''polynomial'' (Xmlt.options [
    (''coefficient'', Xmlt.singleton ''coefficient'' xml2coeff PNum), 
    (''variable'', Xmlt.change (Xmlt.nat ''variable'') (\<lambda>n. PVar (n - 1))),
    (''sum'', Xmlt.many ''sum'' (polynomial xml2coeff) PSum),
    (''product'', Xmlt.many ''product'' (polynomial xml2coeff) PMult)
  ]) id x"

definition int_coeff :: "int xmlt" where
  "int_coeff \<equiv> Xmlt.int ''integer''"
hide_const (open) int_coeff

definition vec_coeff :: "'a xmlt \<Rightarrow> 'a vec xmlt" where
  "vec_coeff xml2coeff \<equiv> Xmlt.many ''vector''
     (Xmlt.singleton ''coefficient'' xml2coeff id) vec_of_list"
hide_const (open) vec_coeff

(* allow vectors and convert them to matrices by filling in zeros *)
definition mat_coeff :: "nat \<Rightarrow> 'a \<Rightarrow> 'a xmlt \<Rightarrow> 'a mat xmlt" where
  "mat_coeff n ze xml2coeff \<equiv> Xmlt.options [
     (''matrix'', Xmlt.many ''matrix'' (CPF_Parser.vec_coeff xml2coeff) (mat_of_cols n)),
     (''vector'', Xmlt.change
       (CPF_Parser.vec_coeff xml2coeff)
       (\<lambda>v. mat n n (\<lambda> (i,j). if j = 0 then v $ i else ze))
     )]"
hide_const (open) mat_coeff

definition arctic_coeff :: "arctic xmlt" where
  "arctic_coeff \<equiv> Xmlt.options [
    (''integer'', Xmlt.change (Xmlt.int ''integer'') Num_arc),
    (''minusInfinity'', Xmlt.leaf ''minusInfinity'' MinInfty)
  ]"
hide_const (open) arctic_coeff

definition arctic_rat_coeff :: "rat arctic_delta xmlt" where
  "arctic_rat_coeff \<equiv> Xmlt.options [
    (''integer'', Xmlt.change Xmlt.rat Num_arc_delta),
    (''rational'', Xmlt.change Xmlt.rat Num_arc_delta),
    (''minusInfinity'', Xmlt.leaf ''minusInfinity'' MinInfty_delta)
  ]"
hide_const (open) arctic_rat_coeff

datatype Domain =
  Natural nat | Integer | Arctic | Arctic_rat | Int_mat nat nat | Arctic_mat nat
| Arctic_rat_mat nat | Rational rat nat | Rat_mat nat nat | Mini_Alg real nat 
| Mini_Alg_mat nat nat

definition rat_domain :: "rat xmlt" where
  "rat_domain \<equiv> Xmlt.singleton ''rationals'' (Xmlt.singleton ''delta'' Xmlt.rat id) id"
hide_const (open) rat_domain

definition real :: "real xmlt" where
  "real x = (try
    Xmlt.change Xmlt.rat (real_of_rat) x
  catch (\<lambda>_. 
    Xmlt.triple ''algebraic'' (Xmlt.rat) (Xmlt.rat) (Xmlt.rat)
      (\<lambda>a b c. (real_of_rat a + real_of_rat b * sqrt (real_of_rat c))) x))"

definition real_domain :: "real xmlt" where
  "real_domain = Xmlt.singleton ''algebraicNumbers'' (Xmlt.singleton ''delta'' real id) id"
hide_const (open) real_domain

definition basic_domain :: "(nat \<Rightarrow> Domain) xmlt" where
  "basic_domain \<equiv> Xmlt.options [
    (''naturals'', Xmlt.leaf ''naturals'' Natural),
    (''integers'', Xmlt.leaf ''integers'' (\<lambda> d. Integer)),
    (''arctic'', Xmlt.singleton ''arctic'' 
      (Xmlt.singleton ''domain'' (Xmlt.options [
        (''naturals'', Xmlt.leaf ''naturals'' (\<lambda> d. Arctic)),
        (''integers'', Xmlt.leaf ''integers'' (\<lambda> d. Arctic)),
        (''rationals'', Xmlt.change CPF_Parser.rat_domain (\<lambda> r d. Arctic_rat))
      ]) id) id),
    (''rationals'', Xmlt.change CPF_Parser.rat_domain Rational),
    (''algebraicNumbers'', Xmlt.change CPF_Parser.real_domain Mini_Alg)
  ]"
hide_const (open) basic_domain

definition interpretation_type :: "Domain xmlt" where
  "interpretation_type \<equiv> (Xmlt.singleton ''type'' 
     (Xmlt.options [
       (''polynomial'', (Xmlt.pair ''polynomial'' 
       (Xmlt.singleton ''domain'' 
       (Xmlt.options [
         (''naturals'', CPF_Parser.basic_domain),
         (''integers'', CPF_Parser.basic_domain),
         (''arctic'', CPF_Parser.basic_domain),
         (''rationals'', CPF_Parser.basic_domain),
         (''algebraicNumbers'', CPF_Parser.basic_domain),
         (''matrices'',Xmlt.triple ''matrices''
           (Xmlt.nat ''dimension'') 
           (Xmlt.nat ''strictDimension'') 
           (Xmlt.singleton ''domain'' CPF_Parser.basic_domain id)
           (\<lambda>di sd domain d. case domain d of
             Natural _ \<Rightarrow> Int_mat di sd
           | Rational _ _ \<Rightarrow> Rat_mat di sd
           | Mini_Alg _ _ \<Rightarrow> Mini_Alg_mat di sd
           | Integer \<Rightarrow> Int_mat di sd
           | Arctic \<Rightarrow> Arctic_mat di
           | Arctic_rat \<Rightarrow> Arctic_rat_mat di
         ))
       ]) 
     id)
     (Xmlt.nat ''degree'') 
     (\<lambda>type d. type d))),
     (''matrixInterpretation'',Xmlt.triple ''matrixInterpretation''
       (Xmlt.singleton ''domain'' CPF_Parser.basic_domain id)
       (Xmlt.nat ''dimension'') 
       (Xmlt.nat ''strictDimension'') 
       (\<lambda>domain di sd. case domain 0 of
         Natural _ \<Rightarrow> Int_mat di sd
       | Integer \<Rightarrow> Int_mat di sd
       | Rational _ _  \<Rightarrow> Rat_mat di sd
       | Mini_Alg _ _  \<Rightarrow> Mini_Alg_mat di sd
       | Arctic \<Rightarrow> Arctic_mat di
       | Arctic_rat \<Rightarrow> Arctic_rat_mat di
     ))
   ])
 id)"
hide_const (open) interpretation_type

type_synonym ('f,'a) inter = "('f \<times> nat) \<times> 'a \<times> 'a list"
type_synonym ('f,'a) mat_inter = "('f \<times> nat) \<times> 'a mat \<times> 'a mat list"
type_synonym ('f,'a) nl_inter = "('f \<times> nat) \<times> ((nat \<times> nat) list \<times> 'a)list"

datatype ('f) "interpretation" = 
  (* linear polynomials (with support of negative constants *)
  Int_linear_poly (int_linear_poly: "('f,int)inter")
| Rat_linear_poly  (rat_linear_poly: "('f,rat)inter")
| Arctic_linear_poly (arctic_linear_poly:  "('f,arctic)inter")
| Arctic_rat_linear_poly  (arctic_rat_linear_poly: "('f,rat arctic_delta)inter")
| Real_linear_poly  (real_linear_poly: "('f,real)inter")
  (* linear polynomials over matrices (with support of negative constants *)
| Int_matrix  (int_matrix: "('f,int)mat_inter")
| Rat_matrix  (rat_matrix: "('f,rat)mat_inter")
| Arctic_matrix  (arctic_matrix: "('f,arctic)mat_inter")
| Arctic_rat_matrix  (arctic_rat_matrix: "('f,rat arctic_delta)mat_inter")
| Real_matrix  (real_matrix: "('f,real)mat_inter")
  (* non linear polynomials (without support of negative constants) *)
| Int_non_linear_poly  (int_non_linear_poly: "('f,int)nl_inter")
| Rat_non_linear_poly  (rat_non_linear_poly: "('f,rat)nl_inter")
| Real_non_linear_poly  (real_non_linear_poly: "('f,real)nl_inter")

(* the flag bi is to indicate that we are interested in bounded increase,
   where rationals always have to be chosen as non-linear variant *) 
abbreviation class_ring where 
  "class_ring t \<equiv> class_semiring t ()"

abbreviation mat_ring :: "('a :: semiring_1) itself \<Rightarrow> nat \<Rightarrow> 'a mat ring" where
  "mat_ring t n \<equiv> ring\<^sub>m t n ()"

definition "interpretation" :: "bool \<Rightarrow> 'a xmlt \<Rightarrow> 'a redtriple_impl xmlt" where
  "interpretation bi xml2name \<equiv> 
      (Xmlt.many1_gen ''interpretation'' 
       CPF_Parser.interpretation_type
       (\<lambda>type. case type of
          Natural deg \<Rightarrow> if deg \<le> 1 \<and> \<not> bi then (\<lambda>c. (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial CPF_Parser.int_coeff x \<guillemotright>= lpoly_of c)
                        (\<lambda>f a poly.  (Int_linear_poly ((f,a),(fst poly, fit_length c a (snd poly)))))))
                   (class_ring (TYPE(int))) else (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial CPF_Parser.int_coeff x \<guillemotright>= (\<lambda> p. return (poly_of p)))
                        (\<lambda>f a poly.  (Int_non_linear_poly ((f,a),poly))))
        | Integer \<Rightarrow> Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial CPF_Parser.int_coeff x \<guillemotright>= (\<lambda> p. return (poly_of p)))
                        (\<lambda>f a poly.  (Int_non_linear_poly ((f,a),poly)))
        | Rational d deg \<Rightarrow> if deg \<le> 1 \<and> \<not> bi then (\<lambda> c. (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial Xmlt.rat x \<guillemotright>= lpoly_of c)
                        (\<lambda>f a poly.  (Rat_linear_poly ((f,a),(fst poly, fit_length c a (snd poly))))))) 
                    (class_ring (TYPE(rat))) else (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial Xmlt.rat x \<guillemotright>= (\<lambda> p. return (poly_of p)))
                        (\<lambda>f a poly.  (Rat_non_linear_poly ((f,a),poly))))
        | Mini_Alg d deg \<Rightarrow> if deg \<le> 1 \<and> \<not> bi then (\<lambda> c. (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial real x \<guillemotright>= lpoly_of c)
                        (\<lambda>f a poly.  (Real_linear_poly ((f,a),(fst poly, fit_length c a (snd poly))))))) 
                    (class_ring (TYPE(real))) else (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial real x \<guillemotright>= (\<lambda> p. return (poly_of p)))
                        (\<lambda>f a poly.  (Real_non_linear_poly ((f,a),poly))))
        | Arctic \<Rightarrow> (\<lambda> c. (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial CPF_Parser.arctic_coeff x \<guillemotright>= lpoly_of c)
                        (\<lambda>f a poly.  (Arctic_linear_poly ((f,a),(fst poly, fit_length c a (snd poly))))))) 
                    (class_ring (TYPE(arctic)))
        | Arctic_rat \<Rightarrow> (\<lambda> c. (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial CPF_Parser.arctic_rat_coeff x \<guillemotright>= lpoly_of c)
                        (\<lambda>f a poly.  (Arctic_rat_linear_poly ((f,a),(fst poly, fit_length c a (snd poly))))))) 
                    (class_ring (TYPE(rat arctic_delta)))
        | Int_mat n sd \<Rightarrow> (\<lambda> c. (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial (CPF_Parser.mat_coeff n 0 CPF_Parser.int_coeff) x \<guillemotright>= lpoly_of c)
                        (\<lambda>f a poly.  (Int_matrix ((f,a),(fst poly, fit_length c a (snd poly))))))) 
                    (mat_ring (TYPE(int)) n)
        | Rat_mat n sd \<Rightarrow> (\<lambda>c. (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial (CPF_Parser.mat_coeff n 0 Xmlt.rat) x \<guillemotright>= lpoly_of c)
                        (\<lambda>f a poly.  (Rat_matrix ((f,a),(fst poly, fit_length c a (snd poly))))))) 
                    (mat_ring (TYPE(rat)) n)
        | Mini_Alg_mat n sd \<Rightarrow> (\<lambda>c. (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial (CPF_Parser.mat_coeff n 0 real) x \<guillemotright>= lpoly_of c)
                        (\<lambda>f a poly.  (Real_matrix ((f,a),(fst poly, fit_length c a (snd poly))))))) 
                    (mat_ring (TYPE(real)) n)
        | Arctic_mat n \<Rightarrow> (\<lambda>c. (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial (CPF_Parser.mat_coeff n MinInfty CPF_Parser.arctic_coeff) x \<guillemotright>= lpoly_of c)
                        (\<lambda> f a poly. (Arctic_matrix ((f,a),(fst poly, fit_length c a (snd poly))))))) 
                    (mat_ring (TYPE(arctic)) n)
        | Arctic_rat_mat n \<Rightarrow> (\<lambda>c. (Xmlt.triple ''interpret''
                        xml2name
                        (Xmlt.nat ''arity'')
                        (\<lambda>x. CPF_Parser.polynomial (CPF_Parser.mat_coeff n MinInfty_delta CPF_Parser.arctic_rat_coeff) x \<guillemotright>= lpoly_of c)
                        (\<lambda>f a poly. (Arctic_rat_matrix ((f,a),(fst poly, fit_length c a (snd poly))))))) 
                    (mat_ring (TYPE(rat arctic_delta)) n)
        )
      (\<lambda> type pi. case type of 
            Natural deg \<Rightarrow> if deg \<le> 1 \<and> \<not> bi then Int_carrier (map int_linear_poly pi) else Int_nl_carrier (map int_non_linear_poly pi)
          | Integer \<Rightarrow> Int_nl_carrier (map int_non_linear_poly pi)
          | Rational d deg \<Rightarrow> if deg \<le> 1 \<and> \<not> bi then Rat_carrier (map rat_linear_poly pi) else Rat_nl_carrier d (map rat_non_linear_poly pi)
          | Mini_Alg d deg \<Rightarrow> if deg \<le> 1 \<and> \<not> bi then Real_carrier (map real_linear_poly pi) else Real_nl_carrier d (map real_non_linear_poly pi)
          | Arctic \<Rightarrow> Arctic_carrier (map arctic_linear_poly pi)
          | Arctic_rat \<Rightarrow> Arctic_rat_carrier (map arctic_rat_linear_poly pi)
          | Int_mat n sd \<Rightarrow> Int_mat_carrier n sd (map int_matrix pi)
          | Rat_mat n sd \<Rightarrow> Rat_mat_carrier n sd (map rat_matrix pi)
          | Mini_Alg_mat n sd \<Rightarrow> Real_mat_carrier n sd (map real_matrix pi)
          | Arctic_mat n \<Rightarrow> Arctic_mat_carrier n (map arctic_matrix pi)
          | Arctic_rat_mat n \<Rightarrow> Arctic_rat_mat_carrier n (map arctic_rat_matrix pi)
      ))
    "

hide_const (open) "interpretation"

definition status_precedence :: "'a xmlt \<Rightarrow> ('a status_prec_repr) xmlt" where
  "status_precedence xml2name \<equiv> Xmlt.many ''statusPrecedence'' 
  (Xmlt.tuple4 ''statusPrecedenceEntry''
      xml2name
      (Xmlt.nat ''arity'')
      (Xmlt.nat ''precedence'')
      (Xmlt.options [(''lex'', Xmlt.leaf ''lex'' Lex), (''mul'', Xmlt.leaf ''mul'' Mul)])
      (\<lambda>f a p s. (((f,a), (p,s))))
  ) id" 
hide_const (open) status_precedence

definition path_order :: "'a xmlt \<Rightarrow> 'a redtriple_impl xmlt" where
  "path_order xml2name \<equiv> xml1to2elements ''pathOrder''
   (CPF_Parser.status_precedence xml2name)
   (CPF_Parser.afs xml2name)
   (\<lambda>prec\<tau> afo. let afs = (case afo of None \<Rightarrow> [] | Some af \<Rightarrow> af)
 in
              RPO prec\<tau> afs)"
hide_const (open) path_order

definition precedence_weight :: "'a xmlt \<Rightarrow> (nat \<Rightarrow> 'a prec_weight_repr) xmlt" where
  "precedence_weight xml2name \<equiv> Xmlt.many ''precedenceWeight'' 
  (Xmlt.guard (\<lambda> x. Xml.num_children x = 4)
    (Xmlt.tuple4 ''precedenceWeightEntry''
        xml2name
        (Xmlt.nat ''arity'')
        (Xmlt.nat ''precedence'')
        (Xmlt.nat ''weight'')
        (\<lambda>f a p w. (((f,a), (p,w,None)))))
    (Xmlt.tuple5 ''precedenceWeightEntry''
        xml2name
        (Xmlt.nat ''arity'')
        (Xmlt.nat ''precedence'')
        (Xmlt.nat ''weight'')
        (Xmlt.many ''subtermCoefficientEntries''
          (Xmlt.nat ''entry'')
          id)
        (\<lambda>f a p w e. (((f,a), (p,w,Some e)))))
  ) Pair" 
hide_const (open) precedence_weight

definition wpo_params :: "'a xmlt \<Rightarrow> 'a wpo_params xmlt" where
  "wpo_params xml2name \<equiv> Xmlt.many ''precedenceStatus'' 
    (Xmlt.tuple4 ''precedenceStatusEntry''
        xml2name
        (Xmlt.nat ''arity'')
        (Xmlt.nat ''precedence'')
        (Xmlt.many ''status'' CPF_Parser.position id)
        (\<lambda>f a p s. (((f,a), (p,s))))
  ) id" 

hide_const (open) wpo_params

definition knuth_bendix_order :: "'a xmlt \<Rightarrow> 'a redtriple_impl xmlt" where
  "knuth_bendix_order xml2name \<equiv> xml2to3elements ''knuthBendixOrder''
   (Xmlt.nat ''w0'')
   (CPF_Parser.precedence_weight xml2name)
   (CPF_Parser.afs xml2name)
   (\<lambda> w0 prw afo. let afs = (case afo of None \<Rightarrow> [] | Some af \<Rightarrow> af)
 in
              KBO (prw w0) afs)"
hide_const (open) knuth_bendix_order


partial_function (sum_bot) redtriple :: "bool \<Rightarrow> 'a xmlt \<Rightarrow> 'a redtriple_impl xmlt" where
  [code]: "redtriple bi xml2name x = 
     Xmlt.singleton ''redPair'' 
     (Xmlt.options [
  (''interpretation'', CPF_Parser.interpretation bi xml2name), 
  (''pathOrder'', CPF_Parser.path_order xml2name),
  (''knuthBendixOrder'', CPF_Parser.knuth_bendix_order xml2name),
  (''weightedPathOrder'', Xmlt.pair ''weightedPathOrder''
      (CPF_Parser.wpo_params xml2name)
      (redtriple bi xml2name)
      WPO)
  ]) id x"
hide_const (open) redtriple

definition level_mapping :: "'a xmlt \<Rightarrow> 'a scnp_af xmlt" where
  "level_mapping xml2name = Xmlt.many ''levelMapping'' 
     (Xmlt.many2 ''levelMappingEntry''
      xml2name
      (Xmlt.nat ''arity'')
      (Xmlt.pair ''positionLevelEntry''
         (Xmlt.nat ''position'')
         (Xmlt.nat ''level'')
         Pair
      )
      (\<lambda>f a ps. ((f,a), map (\<lambda> (p,l). (if p = 0 then a else p - 1,l)) ps))
  ) id" 

definition root_redtriple :: "'a xmlt \<Rightarrow> 'a root_redtriple_impl xmlt" where
  "root_redtriple xml2name = Xmlt.singleton ''redPair'' (       
    Xmlt.triple ''scnp'' 
      (Xmlt.singleton ''status'' 
         (Xmlt.options [
             (''ms'',Xmlt.leaf ''ms'' MS_Ext),
             (''min'',Xmlt.leaf ''min'' Min_Ext),
             (''dms'',Xmlt.leaf ''dms'' Dms_Ext),
             (''max'',Xmlt.leaf ''max'' Max_Ext)
                      ])
         id)     
      (level_mapping xml2name)
      (CPF_Parser.redtriple False xml2name) 
      SCNP) id"
hide_const (open) root_redtriple
      

definition
  ordering_constraint_proof :: "bool \<Rightarrow> 'a xmlt \<Rightarrow> 'a redtriple_impl xmlt"
where
  "ordering_constraint_proof bi xml2name = 
     Xmlt.singleton ''orderingConstraintProof'' (CPF_Parser.redtriple bi xml2name) id"

definition
  root_ordering_constraint_proof :: "'a xmlt \<Rightarrow> 'a root_redtriple_impl xmlt"
where
  "root_ordering_constraint_proof xml2name = 
     Xmlt.singleton ''orderingConstraintProof'' (CPF_Parser.root_redtriple xml2name) id"

definition
  ordering_constraint_proof' :: "bool \<Rightarrow> 'a xmlt \<Rightarrow> ('a root_redtriple_impl + 'a redtriple_impl) xmlt"
where
  "ordering_constraint_proof' bi xml2name = Xmlt.choice ''reduction pair''
     [Xmlt.change (ordering_constraint_proof bi xml2name) Inr,
      Xmlt.change (root_ordering_constraint_proof xml2name) Inl]"

hide_const (open) ordering_constraint_proof
hide_const (open) root_ordering_constraint_proof
hide_const (open) ordering_constraint_proof'

definition
  id_labeling :: "(('f,'a list)lab,'a,'a list + ('f,'a list)lab list)label" 
where
  "id_labeling _ cs = Inl cs" 

definition scg_position :: "nat xmlt" where
  "scg_position \<equiv> Xmlt.nat ''position''"
hide_const (open) scg_position

definition
  scg :: "'f xmlt \<Rightarrow> (('f, string) rule \<times> ((nat \<times> nat) list \<times> (nat \<times> nat) list)) xmlt"
where
  "scg xml2name \<equiv> Xmlt.many1 ''sizeChangeGraph'' 
    (CPF_Parser.rule xml2name)
    (Xmlt.triple ''edge''
       CPF_Parser.scg_position
       (Xmlt.bool ''strict'')
       CPF_Parser.scg_position
       (\<lambda>p s q. (s,(p,q)))
    )
    (\<lambda>lr edges. (lr,(map snd (filter fst edges), map snd (filter (\<lambda> x. \<not> fst x) edges))))"
hide_const (open) scg

definition uncurry_info :: "'f xmlt \<Rightarrow> ('f, string) uncurry_info xmlt" where
  "uncurry_info xml2name \<equiv> Xmlt.tuple4 ''uncurryInformation''
    xml2name
    (Xmlt.many ''uncurriedSymbols''
      (Xmlt.many2 ''uncurriedSymbolEntry''
        xml2name
        (Xmlt.nat ''arity'')
        xml2name
        (\<lambda>f n fs. ((f, n), fs))) id)
    (Xmlt.singleton ''uncurryRules'' (CPF_Parser.rules xml2name) id) 
    (Xmlt.singleton ''etaRules'' (CPF_Parser.rules xml2name) id) 
    (\<lambda>a sml U E. (a, sml, U, E))"

definition innermostLhss :: "'a xmlt \<Rightarrow> ('a, string) term list xmlt"
where
  "innermostLhss xml2name = Xmlt.many ''innermostLhss'' (CPF_Parser.term xml2name) id"

definition strategy :: "'a xmlt \<Rightarrow> ('a,string)strategy xmlt" where
  "strategy xml2name = Xmlt.singleton ''strategy'' 
     (Xmlt.options [
        (''innermost'', Xmlt.leaf ''innermost'' Innermost),
        (''innermostLhss'', Xmlt.change (innermostLhss xml2name) Innermost_Q)
                  ]) id"


definition
  xml2dp_input :: "'f xmlt \<Rightarrow> (bool \<times> ('f, string) rules \<times> ('f,string)strategy \<times> ('f, string) rules) xmlt"
where
  "xml2dp_input xml2name \<equiv> Xmlt.guard (\<lambda> x. Xml.num_children x = 4)
   (Xmlt.tuple4 ''dpInput'' 
         (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
         (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id)
         (strategy xml2name)  
         (Xmlt.bool ''minimal'')
         (\<lambda> r p s m. (m,p,s,r)))
   (Xmlt.triple ''dpInput'' 
         (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
         (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id)
         (Xmlt.bool ''minimal'')
         (\<lambda> r p m. (m,p,No_Strategy,r)))"

definition
  xml2dp_input' :: "bool \<Rightarrow> ('f,'l)lab xmlt \<Rightarrow> ('f,'l,string)dppLL xmlt"
where
  "xml2dp_input' termination xml2name \<equiv> Xmlt.change (xml2dp_input xml2name) 
    (\<lambda> (m,p,q,r). (if termination then default_nfs_dp else default_nfs_nt_dp, m, p, [], strategy_to_Q q r, [], r))"

partial_function (sum_bot) 
  xml2cond_constraint :: "'f xmlt \<Rightarrow> ('f,string) cond_constraint xmlt" where
  [code]: "xml2cond_constraint xml2name x = Xmlt.singleton ''conditionalConstraint''
    (Xmlt.options [
      (''all'',Xmlt.pair ''all'' (Xmlt.text ''var'') (xml2cond_constraint xml2name) CC_all),
      (''implication'',Xmlt.many1 ''implication'' (xml2cond_constraint xml2name) (xml2cond_constraint xml2name) 
        (\<lambda> c cs. let ccs = c # cs in CC_impl (take (length cs) ccs) (last ccs))),
      (''constraint'',Xmlt.triple ''constraint'' 
        (CPF_Parser.term xml2name) 
        (Xmlt.options [
          (''rewrite'',Xmlt.leaf ''rewrite'' None),
          (''strict'',Xmlt.leaf ''strict'' (Some True)),
          (''nonStrict'',Xmlt.leaf ''nonStrict'' (Some False))
          ]) 
        (CPF_Parser.term xml2name) 
        (\<lambda> s rel t. case rel of None \<Rightarrow> CC_rewr s t | Some stri \<Rightarrow> CC_cond stri (s,t)))
    ]) id x"

partial_function (sum_bot) xml2cond_constraint_prf :: "'f xmlt \<Rightarrow> ('f,string) cond_constraint_prf xmlt" where
  [code]: "xml2cond_constraint_prf xml2name x = (let cc = xml2cond_constraint xml2name in Xmlt.singleton ''conditionalConstraintProof''
    (Xmlt.options [
      (''final'',Xmlt.leaf ''final'' Final),
      (''differentConstructor'', Xmlt.singleton ''differentConstructor'' cc Different_Constructor),
      (''sameConstructor'', Xmlt.triple ''sameConstructor'' cc cc (xml2cond_constraint_prf xml2name) Same_Constructor),
      (''variableEquation'', Xmlt.tuple4 ''variableEquation'' (Xmlt.text ''var'') (CPF_Parser.term xml2name) cc (xml2cond_constraint_prf xml2name) Variable_Equation),
      (''funargIntoVar'', Xmlt.tuple5 ''funargIntoVar'' cc CPF_Parser.position (Xmlt.text ''var'') cc (xml2cond_constraint_prf xml2name) Funarg_Into_Var),
      (''simplifyCondition'', Xmlt.tuple4 ''simplifyCondition'' cc (CPF_Parser.subst xml2name) cc (xml2cond_constraint_prf xml2name) Simplify_Condition),
      (''induction'', Xmlt.triple ''induction'' cc (Xmlt.many ''conjuncts'' cc id) (Xmlt.many ''ruleConstraintProofs'' (
        Xmlt.tuple4 ''ruleConstraintProof'' 
          (CPF_Parser.rule xml2name)
          (Xmlt.many ''subtermVarEntries''
            (Xmlt.many1 ''subtermVarEntry''
              (CPF_Parser.term xml2name)
              (Xmlt.text ''var'')
              Pair)
            id)
          cc
          (xml2cond_constraint_prf xml2name)
          (\<lambda> lr rys cc p. (lr,rys,cc,p))
        ) id) Induction),
      (''deleteCondition'', Xmlt.pair ''deleteCondition'' cc(xml2cond_constraint_prf xml2name) Delete_Condition)
    ])
    id
    x)"

definition
  xml2cond_red_pair_proof :: "'f xmlt \<Rightarrow> ('f, string) cond_red_pair_prf xmlt" where
  "xml2cond_red_pair_proof xml2name \<equiv> Xmlt.tuple4 ''condRedPairProof'' 
     xml2name
     (Xmlt.nat ''before'')
     (Xmlt.nat ''after'')
     (Xmlt.many ''conditions''
       (Xmlt.triple ''condition''
         (xml2cond_constraint xml2name)
         (Xmlt.singleton ''dpSequence'' (CPF_Parser.rules xml2name) id)
         (xml2cond_constraint_prf xml2name)
         (\<lambda> c s p. (c,s,p)))
       id)
     (\<lambda> c b a ccs. Cond_Red_Pair_Prf c ccs b a)"

definition xmlIgnore :: "string \<Rightarrow> unit xmlt" where
  "xmlIgnore name xml \<equiv> if name = (Xml.tag xml)
    then return ()
    else error (''expected '' @ name @ '' element, but found '' @ show xml)"

definition bounds_type :: "boundstype xmlt" where
  "bounds_type \<equiv> Xmlt.singleton ''type'' 
     (Xmlt.options [
         (''roof'', Xmlt.leaf ''roof'' Roof),
         (''match'', Xmlt.leaf ''match'' Match)
     ])
     id"

definition bounds_bound :: "nat xmlt" where
  "bounds_bound \<equiv> Xmlt.nat ''bound''"

definition state :: "string xmlt" where
  "state = Xmlt.text ''state''"

definition final_states :: "string list xmlt" where
  "final_states = Xmlt.many ''finalStates'' state id"

definition ta_bounds_lhs :: "'a xmlt \<Rightarrow> (('a \<times> nat) \<times> string list)xmlt" where
  "ta_bounds_lhs xml2name = (Xmlt.many2 ''lhs''
      (xml2name)
      (Xmlt.nat ''height'')
      (state)
      (\<lambda>f h qs. ((f, h), qs)))"

definition ta_normal_lhs :: "'a xmlt \<Rightarrow> ('a \<times> string list)xmlt" where
  "ta_normal_lhs xml2name = (Xmlt.many1 ''lhs''
      (xml2name)
      (state)
      Pair)"

definition transition_lhs :: "'a xmlt \<Rightarrow> ('a + string) xmlt" where
  "transition_lhs xml2lhs = Xmlt.choice ''lhs'' [ 
    (Xmlt.singleton ''lhs'' state Inr),
    (Xmlt.change xml2lhs Inl)]"

definition
  transition :: "('a \<times> string list) xmlt \<Rightarrow> ((string, 'a) ta_rule + (string \<times> string)) xmlt"
where
  "transition xml2lhs = Xmlt.pair ''transition''
    (transition_lhs xml2lhs)
    (Xmlt.singleton ''rhs'' state id)
    (\<lambda>lhs q. case lhs of
      Inl (f, qs) \<Rightarrow> Inl (TA_rule f qs q)
    | Inr q' \<Rightarrow> Inr (q', q))"

definition
  transitions ::
    "('a \<times> string list) xmlt \<Rightarrow> (((string, 'a) ta_rule list) \<times> ((string \<times> string) list)) xmlt"
where
  "transitions xml2lhs = Xmlt.many ''transitions'' (transition xml2lhs)
     (\<lambda>rls. let (rules, eps) = partition (\<lambda>rl. case rl of Inl _ \<Rightarrow> True | Inr _ \<Rightarrow> False) rls;
                        ruls = map (\<lambda>rl. case rl of Inl r \<Rightarrow> r) rules;
                        ep   = map (\<lambda>rl. case rl of Inr e \<Rightarrow> e) eps
            in (ruls, ep))"

definition tree_automaton :: "('a \<times> string list) xmlt \<Rightarrow> (string, 'a) tree_automaton xmlt" where
  "tree_automaton xml2lhs \<equiv> Xmlt.pair ''treeAutomaton''
    final_states
    (transitions xml2lhs)
    (\<lambda> f (t, eps). Tree_Automaton f t eps)"

definition closed_criterion :: "string ta_relation xmlt" where
  "closed_criterion = Xmlt.singleton ''criterion'' (Xmlt.options [
      (''compatibility'', Xmlt.leaf ''compatibility'' Id_Relation),
      (''stateCompatibility'', Xmlt.singleton ''stateCompatibility'' 
         (Xmlt.many ''relation''
             (Xmlt.pair ''entry'' state state Pair)
             id) Some_Relation),
      (''decisionProcedure'', Xmlt.leaf ''decisionProcedure'' Decision_Proc)
    ]) id"

definition bounds_info :: "'a xmlt \<Rightarrow> ('a, string) bounds_info xmlt" where
  "bounds_info xml2name \<equiv> Xmlt.guard (\<lambda> x. Xml.num_children x = 4) 
    (Xmlt.tuple4 ''bounds''
      bounds_type
      bounds_bound
      final_states
      (tree_automaton (ta_bounds_lhs xml2name))
      (\<lambda> a b c d. Bounds_Info a b c d Id_Relation))
    (Xmlt.tuple5 ''bounds''
      bounds_type
      bounds_bound
      final_states
      (tree_automaton (ta_bounds_lhs xml2name))
      closed_criterion  
      Bounds_Info)
    "

definition 
  crule :: "'f xmlt \<Rightarrow> ('f,string)crule xmlt"
where
  "crule xml2name \<equiv> Xmlt.many1 ''conditionalRule'' (CPF_Parser.rule xml2name) (CPF_Parser.rule xml2name) Pair"

definition 
  crules :: "'f xmlt \<Rightarrow> ('f,string)crules xmlt"
where
  "crules xml2name \<equiv> Xmlt.many ''conditionalRules'' (crule xml2name) id"

definition 
  xml2ctrs_input :: "'f xmlt \<Rightarrow> ('f,string)crules xmlt"
where 
  "xml2ctrs_input xml2name \<equiv> Xmlt.singleton ''ctrsInput'' (crules xml2name) id"

definition
  xml2completion_input :: "'f xmlt \<Rightarrow> (('f, string) equation list \<times> ('f, string) rules) xmlt"
where
  "xml2completion_input xml2name \<equiv> 
     Xmlt.pair ''completionInput''
       (Xmlt.singleton ''equations'' (CPF_Parser.rules xml2name) id)
       (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
       Pair"

definition
  xml2equational_input :: "'f xmlt \<Rightarrow> (('f, string) equation list \<times> ('f, string) term \<times> ('f, string) term) xmlt"
where
  "xml2equational_input xml2name \<equiv>
    Xmlt.pair ''equationalReasoningInput''
      (Xmlt.singleton ''equations'' (CPF_Parser.rules xml2name) id)
      (Xmlt.pair ''equation''
        (CPF_Parser.term xml2name)
        (CPF_Parser.term xml2name)
        (Pair))
      Pair"

definition complexity_class :: "complexity_class xmlt"
where
  "complexity_class = Xmlt.change (Xmlt.nat ''polynomial'') Comp_Poly"

definition symbols :: "string \<Rightarrow> 'f xmlt \<Rightarrow> ('f \<times> nat) list xmlt"
where
  "symbols tagname xml2name =
    Xmlt.many tagname (Xmlt.pair ''symbol'' xml2name (Xmlt.nat ''arity'') Pair) id"

definition "signature = symbols ''signature''"

definition complexity_measure :: "'f xmlt \<Rightarrow> ('f,string)complexity_measure xmlt"
  where "complexity_measure xml2name \<equiv> Xmlt.options [
   (''derivationalComplexity'', Xmlt.singleton ''derivationalComplexity'' (signature xml2name) Derivational_Complexity),
   (''runtimeComplexity'', Xmlt.pair ''runtimeComplexity'' (signature xml2name) (signature xml2name) Runtime_Complexity)
  ]"

definition renaming :: "'f xmlt \<Rightarrow> ('f \<times> 'f)list xmlt" where
  "renaming xml2name \<equiv> Xmlt.many ''renaming''
    (Xmlt.pair ''renamingEntry''
      xml2name
      xml2name
      Pair
   ) id"

hide_const (open) renaming


definition xml2unknown_input :: "unknown_info xmlt"
  where "xml2unknown_input \<equiv> 
       Xmlt.text ''unknownInput''"

hide_const (open) crule crules

datatype ('f,'v)fp_strategy = 
  Outermost
| Context_Sensitive "'f muL" 
| Forbidden_Patterns "('f,'v)forb_pattern list"

primrec strategy_to_fp :: "('f,string)fp_strategy \<Rightarrow> ('f,string)rules \<Rightarrow> ('f,string)forb_pattern list" where
  "strategy_to_fp (Forbidden_Patterns p) r = p"
| "strategy_to_fp Outermost r = o_to_fp_impl (map fst r)"
| "strategy_to_fp (Context_Sensitive \<mu>) r = mu_to_fp_impl \<mu>"

datatype ('f,'v) input = 
  DP_input "(bool \<times> ('f, 'v) rules \<times> ('f, 'v) strategy \<times> ('f, 'v) rules)"
| Inn_TRS_input "(('f, 'v) strategy \<times> ('f, 'v) rules \<times> ('f, 'v) rules option)"
| COMP_input "(('f, 'v) equation list \<times> ('f, 'v) rules)"
| EQ_input "(('f, 'v) equation list \<times> ('f, 'v) term \<times> ('f, 'v) term)"
| CPX_input "(('f, 'v)strategy \<times> ('f, 'v) rules \<times> ('f, 'v) rules option \<times> ('f,'v)complexity_measure \<times> complexity_class)"
| FP_TRS_input "(('f, 'v) fp_strategy \<times> ('f, 'v) rules)"
| CTRS_input "('f, 'v) crules"
| TA_input "(string,'f)tree_automaton" "('f,'v)rules"
| Unknown_input unknown_info

type_synonym ('f,'v)inn_fp_strategy = "('f,'v)strategy + ('f,'v)fp_strategy"

definition
  replacement_map :: "'a xmlt \<Rightarrow> 'a muL xmlt"
where
  "replacement_map xml2name \<equiv> Xmlt.many ''contextSensitive''
    (Xmlt.many2 ''replacementMapEntry''
       xml2name
       (Xmlt.nat ''arity'')
       CPF_Parser.position
       (\<lambda> f a is. ((f,a),is)))
    id"
hide_const (open) replacement_map

definition inn_fp_strategy :: "'a xmlt \<Rightarrow> ('a,string)inn_fp_strategy xmlt" where
  "inn_fp_strategy xml2name = Xmlt.singleton ''strategy'' 
     (Xmlt.options [
        (''innermost'', Xmlt.leaf ''innermost'' (Inl Innermost)),
        (''outermost'', Xmlt.leaf ''outermost'' (Inr Outermost)),
        (''contextSensitive'', Xmlt.change (CPF_Parser.replacement_map xml2name) (\<lambda> p. Inr (Context_Sensitive p))),
        (''forbiddenPatterns'', Xmlt.change (CPF_Parser.forbidden_patterns xml2name) (\<lambda> p. Inr (Forbidden_Patterns p))),
        (''innermostLhss'', Xmlt.change (innermostLhss xml2name) (\<lambda> q. Inl (Innermost_Q q)))
                  ]) id"

definition
  xml2pre_trs_input :: "'f xmlt \<Rightarrow> ('f,string)input xmlt" 
where
  "xml2pre_trs_input xml2name x \<equiv> do {
     let e = Xmlt.fail ''trsInput'' x;
     res \<leftarrow>
      ((case x of XML tag [] (c # cs) \<Rightarrow> if length cs \<le> 2 \<and> tag = ''trsInput'' then (
      do { 
         r \<leftarrow> (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) c;
         (str,cs') \<leftarrow> (if cs = [] \<or> Xml.tag (hd cs) \<noteq> ''strategy'' then return (Inl No_Strategy, cs) else do { s \<leftarrow> (inn_fp_strategy xml2name) (hd cs); return (s,tl cs)});
         (s_opt,cs'') \<leftarrow> if (cs' = [] \<or> Xml.tag (hd cs') \<noteq> ''relativeRules'') then return (None, cs') else 
           do { s \<leftarrow> (Xmlt.singleton ''relativeRules'' (CPF_Parser.rules xml2name) id) (hd cs'); return (Some s,tl cs')};
         if cs'' = [] then return (str,r,s_opt) else e
      }
     ) else e | _ \<Rightarrow> e));
     case res of
       (Inl istrat,r,so) \<Rightarrow> return (Inn_TRS_input (istrat,r,so))
     | (Inr fpstrat,r,None) \<Rightarrow> return (FP_TRS_input (fpstrat,r))
     | _ \<Rightarrow> error (''the combination of relative rules with strategies is only supported for innermost'')}" 


definition
  xml2trs_input :: "'f xmlt \<Rightarrow> (('f,string)strategy \<times> ('f, string) rules \<times> ('f, string) rules option) xmlt"
where
  "xml2trs_input xml2name x \<equiv> do {
    inp <- xml2pre_trs_input xml2name x;
    case inp of Inn_TRS_input i \<Rightarrow> return i
             |  _ \<Rightarrow> error ''require innermost trs at this point''
  }"

definition
  xml2complexity_input :: "'f xmlt \<Rightarrow> (('f,string)strategy \<times> ('f, string) rules \<times> ('f, string) rules option \<times> ('f,string)complexity_measure \<times> complexity_class) xmlt"
where
  "xml2complexity_input xml2name \<equiv> 
     Xmlt.triple ''complexityInput''
       (xml2trs_input xml2name)
       (complexity_measure xml2name)
       complexity_class
       (\<lambda> (q,r,s) cm cc. (q,r,s,cm,cc))"

definition
  xml2complexity_input' :: "('f,'l)lab xmlt \<Rightarrow> ('f,'l,string)complexityLL xmlt"
where
  "xml2complexity_input' xml2name \<equiv> Xmlt.change (xml2complexity_input xml2name)
    (\<lambda> (q,s,w,cm,cc). (strategy_to_Q q (s @ rel_rules_of w), s, rel_rules_of w, cm,cc))"

definition
  xml2input :: "'f xmlt \<Rightarrow> ('f,string) input xmlt"
where
  "xml2input xml2name \<equiv> Xmlt.singleton ''input'' (Xmlt.options  
    [(''dpInput'', Xmlt.change (xml2dp_input xml2name) DP_input),
     (''completionInput'', Xmlt.change (xml2completion_input xml2name) COMP_input),
     (''trsInput'', xml2pre_trs_input xml2name),
     (''equationalReasoningInput'', Xmlt.change (xml2equational_input xml2name) EQ_input),
     (''complexityInput'', Xmlt.change (xml2complexity_input xml2name) CPX_input),
     (''ctrsInput'', Xmlt.change (xml2ctrs_input xml2name) CTRS_input),
     (''treeAutomatonProblem'', Xmlt.pair ''treeAutomatonProblem'' 
        (tree_automaton (ta_normal_lhs xml2name)) (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) TA_input),
     (''unknownInput'', Xmlt.change xml2unknown_input Unknown_input)
     ]) id" 

definition
  xml2inn_fp_trs_assm :: "('f,'l)lab xmlt \<Rightarrow> (('f,'l,string)qreltrsLL + ('f,'l,string)fptrsLL) xmlt" where
  "xml2inn_fp_trs_assm xml2name x \<equiv> do {
    inp <- xml2pre_trs_input xml2name x;
    case inp of 
      Inn_TRS_input (inn,r,so) \<Rightarrow> return (Inl (default_nfs_trs, strategy_to_Q inn r, r, case so of None \<Rightarrow> [] | Some s \<Rightarrow> s))
    | FP_TRS_input (fp,r) \<Rightarrow> return (Inr (strategy_to_fp fp r, r))
    | _ \<Rightarrow> error ''trs as input required''
    }"

definition
  xml2inn_trs_assm :: "('f,'l)lab xmlt \<Rightarrow> ('f,'l,string)qreltrsLL xmlt" where
  "xml2inn_trs_assm xml2name x \<equiv> do {
    inp <- xml2inn_fp_trs_assm xml2name x;
    case inp of
      Inl i \<Rightarrow> return i
    | _ \<Rightarrow> error ''innermost (relative) TRS expected at this point''
   }"

definition
  xml2fp_trs_assm :: "('f,'l)lab xmlt \<Rightarrow> ('f,'l,string)fptrsLL xmlt" where
  "xml2fp_trs_assm xml2name x \<equiv> do {
    inp <- xml2inn_fp_trs_assm xml2name x;
    case inp of
      Inr i \<Rightarrow> return i
    | _ \<Rightarrow> error ''FP TRS expected at this point''
   }"

abbreviation sub_proof :: "(xml \<Rightarrow> string +\<^sub>\<bottom> ('f, 'l) lab)
    \<Rightarrow> (xml \<Rightarrow> string +\<^sub>\<bottom> 'a)
    \<Rightarrow> (xml \<Rightarrow> string +\<^sub>\<bottom> 'b)
    \<Rightarrow> (xml \<Rightarrow> string +\<^sub>\<bottom> 'c)
    \<Rightarrow> (xml \<Rightarrow> string +\<^sub>\<bottom> 'd) \<Rightarrow> xml \<Rightarrow> string +\<^sub>\<bottom> ('f, 'l, string, 'a, 'b, 'c, 'd) assm_proof"
  where "sub_proof xml2name 
  xml2trs_termination_proof 
  xml2dp_termination_proof
  xml2fptrs_termination_proof 
  xml2unknown_proof x \<equiv> 
    let cs = Xml.children x
     in (if length cs \<noteq> 2 then Xmlt.fail ''subProof'' x
      else let inp = take 2 (Xml.tag (cs ! 1)) in
      (if inp = ''tr'' \<or> inp = ''re'' then 
        do {
          io_trs <- xml2inn_fp_trs_assm xml2name (hd cs);
          (case io_trs of
            Inl qtrs \<Rightarrow> Xmlt.change xml2trs_termination_proof (\<lambda> prf. SN_assm_proof qtrs prf :: ('f, 'l, string, 'a, 'b, 'c, 'd) assm_proof)
          | Inr fptrs \<Rightarrow> Xmlt.change xml2fptrs_termination_proof (SN_FP_assm_proof fptrs)) (cs ! 1)
        }
      else if inp = ''dp'' then
        Xmlt.pair ''subProof'' (xml2dp_input' True xml2name) xml2dp_termination_proof Finite_assm_proof x
      else if inp = ''un'' then
        Xmlt.pair ''subProof'' xml2unknown_input xml2unknown_proof Unknown_assm_proof x
      else Xmlt.fail ''subProof'' x))"

abbreviation dp_termination_proof where "dp_termination_proof xml2name 
  xml2trs_termination_proof 
  xml2dp_termination_proof 
  xml2fptrs_termination_proof 
  xml2unknown_proof \<equiv>
  Xmlt.singleton ''dpProof'' (Xmlt.options [ 
    (''pIsEmpty'', Xmlt.leaf ''pIsEmpty'' P_is_Empty),
    (''depGraphProc'', Xmlt.many ''depGraphProc''
      (xml2nd_choice ''component'' 
        (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id)
        ''realScc''
        xml2dp_termination_proof
        (\<lambda>dps prfOpt. (prfOpt, dps)))
      Dep_Graph_Proc),
    (''redPairProc'', Xmlt.triple ''redPairProc'' 
      (CPF_Parser.ordering_constraint_proof' False xml2name) 
      (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id) 
      xml2dp_termination_proof 
      Redpair_Proc),
    (''usableRulesProc'', Xmlt.pair ''usableRulesProc'' 
      (Xmlt.singleton ''usableRules'' (CPF_Parser.rules xml2name) id) 
      xml2dp_termination_proof 
      Usable_Rules_Proc),
    (''innermostLhssRemovalProc'', Xmlt.pair ''innermostLhssRemovalProc'' 
      (innermostLhss xml2name)
      xml2dp_termination_proof 
      Q_Reduction_Proc),
    (''rewritingProc'', Xmlt.guard (\<lambda> x. Xml.num_children x = 3) 
       (Xmlt.triple ''rewritingProc'' 
         (CPF_Parser.rule xml2name)
         (CPF_Parser.rstep xml2name)
         xml2dp_termination_proof 
         (\<lambda> (s,t) (p,lr,t') prof. Rewriting_Proc None (s,t) (s,t') (s,t') lr p prof))
       (Xmlt.guard (\<lambda> x. Xml.num_children x = 4)
         (Xmlt.guard (\<lambda> x. Xml.tag (Xml.children x ! 2) = ''usableRules'')
           (Xmlt.tuple4 ''rewritingProc'' 
             (CPF_Parser.rule xml2name)
             (CPF_Parser.rstep xml2name)
             (Xmlt.singleton ''usableRules'' (CPF_Parser.rules xml2name) id) 
             xml2dp_termination_proof 
             (\<lambda> (s,t) (p,lr,t') U prof. Rewriting_Proc (Some U) (s,t) (s,t') (s,t') lr p prof))
           (Xmlt.tuple4 ''rewritingProc'' 
             (CPF_Parser.rule xml2name)
             (CPF_Parser.rstep xml2name)
             (CPF_Parser.rule xml2name)
             xml2dp_termination_proof 
             (\<lambda> (s,t) (p,lr,t') st'' prof. Rewriting_Proc None (s,t) (s,t') st'' lr p prof)))
       (Xmlt.tuple5 ''rewritingProc'' 
         (CPF_Parser.rule xml2name)
         (CPF_Parser.rstep xml2name)
         (CPF_Parser.rule xml2name)
         (Xmlt.singleton ''usableRules'' (CPF_Parser.rules xml2name) id) 
         xml2dp_termination_proof 
         (\<lambda> (s,t) (p,lr,t') st'' U prof. Rewriting_Proc (Some U) (s,t) (s,t') st'' lr p prof)))),
    (''narrowingProc'', Xmlt.tuple4 ''narrowingProc'' 
      (CPF_Parser.rule xml2name)
      CPF_Parser.pos
      (Xmlt.singleton ''narrowings'' (CPF_Parser.rules xml2name) id) 
      xml2dp_termination_proof 
      Narrowing_Proc),
    (''instantiationProc'', Xmlt.triple ''instantiationProc'' 
      (CPF_Parser.rule xml2name)
      (Xmlt.singleton ''instantiations'' (CPF_Parser.rules xml2name) id) 
      xml2dp_termination_proof 
      Instantiation_Proc),
    (''forwardInstantiationProc'', xml3to4elements ''forwardInstantiationProc'' 
      (CPF_Parser.rule xml2name)
      (Xmlt.singleton ''instantiations'' (CPF_Parser.rules xml2name) id) 
      (Xmlt.singleton ''usableRules'' (CPF_Parser.rules xml2name) id) 
      xml2dp_termination_proof 
      Forward_Instantiation_Proc),
    (''semlabProc'',Xmlt.guard (\<lambda> x. Xml.num_children x = 4) 
     (Xmlt.tuple4 ''semlabProc'' 
       (CPF_Parser.sl_variant xml2name) 
       (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id) 
       (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
       xml2dp_termination_proof
       (\<lambda>sli lp. Semlab_Proc sli lp []))
     (Xmlt.tuple5 ''semlabProc'' 
       (CPF_Parser.sl_variant xml2name) 
       (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id) 
       (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
       (innermostLhss xml2name)
       xml2dp_termination_proof
       (\<lambda>sli lp lr lq. Semlab_Proc sli lp lq lr))),
    (''subtermProc'', xml1many2elements_gen ''subtermProc''
      (CPF_Parser.proj xml2name)
      (\<lambda>\<pi>. CPF_Parser.projected_rseq xml2name (create_proj \<pi>))
      (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id)
      xml2dp_termination_proof
      Subterm_Criterion_Proc),
    (''redPairUrProc'', Xmlt.tuple4 ''redPairUrProc'' 
      (CPF_Parser.ordering_constraint_proof' False xml2name) 
      (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id) 
      (Xmlt.singleton ''usableRules'' (CPF_Parser.rules xml2name) id) 
      xml2dp_termination_proof
      Redpair_UR_Proc),
    (''monoRedPairUrProc'', Xmlt.tuple5 ''monoRedPairUrProc'' 
      (CPF_Parser.ordering_constraint_proof False xml2name) 
      (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id) 
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
      (Xmlt.singleton ''usableRules'' (CPF_Parser.rules xml2name) id) 
      xml2dp_termination_proof
      Mono_Redpair_UR_Proc),
    (''monoRedPairProc'', Xmlt.tuple4 ''monoRedPairProc'' 
      (CPF_Parser.ordering_constraint_proof False xml2name) 
      (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id) 
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
      xml2dp_termination_proof
      Mono_Redpair_Proc),
    (''innermostMonoRedPairProc'', Xmlt.triple ''innermostMonoRedPairProc'' 
      (CPF_Parser.ordering_constraint_proof False xml2name)
      (Xmlt.pair ''deleted'' 
        (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id) 
        (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
        Pair)
      xml2dp_termination_proof
      (\<lambda> rp (p,r). Mono_URM_Redpair_Proc rp p r)),
    (''uncurryProc'', Xmlt.guard (\<lambda> x. Xml.num_children x = 4) 
      (Xmlt.tuple4 ''uncurryProc'' 
        (uncurry_info xml2name) 
        (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id) 
        (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
        xml2dp_termination_proof
        (Uncurry_Proc None))
      (Xmlt.tuple5 ''uncurryProc'' 
        (Xmlt.nat ''applicativeTop'')
        (uncurry_info xml2name) 
        (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id) 
        (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
        xml2dp_termination_proof
        (\<lambda>n. Uncurry_Proc (Some n)))),
    (''flatContextClosureProc'', Xmlt.tuple5 ''flatContextClosureProc''
      (Xmlt.singleton ''freshSymbol'' xml2name id)
      (CPF_Parser.flat_contexts xml2name)
      (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id)
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
      xml2dp_termination_proof
      Fcc_Proc),
    (''switchInnermostProc'', Xmlt.pair ''switchInnermostProc''
      (wcr_proof xml2name)
      xml2dp_termination_proof
      Switch_Innermost_Proc),
    (''splitProc'', Xmlt.tuple4 ''splitProc'' 
      (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id) 
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
      xml2dp_termination_proof
      xml2dp_termination_proof
      Split_Proc),
    (''finitenessAssumption'', Xmlt.singleton ''finitenessAssumption''
      (xml2dp_input' True xml2name)
      (\<lambda> dpp. Assume_Finite dpp [])),
    (''unknownProof'', Xmlt.many2 ''unknownProof''
      (Xmlt.text ''description'')
      (xml2dp_input' True xml2name)
      (sub_proof xml2name
           xml2trs_termination_proof
           xml2dp_termination_proof
           xml2fptrs_termination_proof 
           xml2unknown_proof)
      (\<lambda> desc. Assume_Finite)),
    (''switchToTRS'',Xmlt.singleton ''switchToTRS''
         xml2trs_termination_proof
         To_Trs_Proc),
    (''unlabProc'',Xmlt.triple ''unlabProc''
         (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id) 
         (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
         xml2dp_termination_proof
         Unlab_Proc),
    (''generalRedPairProc'', Xmlt.guard (\<lambda> x. Xml.num_children x = 5) 
      (Xmlt.tuple5 ''generalRedPairProc'' 
      (CPF_Parser.ordering_constraint_proof True xml2name) 
      (Xmlt.singleton ''strict'' (CPF_Parser.rules xml2name) id) 
      (Xmlt.singleton ''bound'' (CPF_Parser.rules xml2name) id)
      (xml2cond_red_pair_proof xml2name)
      xml2dp_termination_proof 
      (\<lambda> rp s b c p. General_Redpair_Proc rp s b c [p]))
      (Xmlt.tuple6 ''generalRedPairProc'' 
      (CPF_Parser.ordering_constraint_proof True xml2name) 
      (Xmlt.singleton ''strict'' (CPF_Parser.rules xml2name) id) 
      (Xmlt.singleton ''bound'' (CPF_Parser.rules xml2name) id)
      (xml2cond_red_pair_proof xml2name)
      xml2dp_termination_proof 
      xml2dp_termination_proof 
      (\<lambda> rp s b c ps pb. General_Redpair_Proc rp s b c [ps,pb]))),
    (''complexConstantRemovalProc'',Xmlt.triple ''complexConstantRemovalProc''
         (CPF_Parser.term xml2name)
         (Xmlt.many ''ruleMap'' 
            (Xmlt.pair ''ruleMapEntry''
              (CPF_Parser.rule xml2name)
              (CPF_Parser.rule xml2name)
              Pair) id) 
         xml2dp_termination_proof
         (\<lambda> t rls prf. Complex_Constant_Removal_Proc (Complex_Constant_Removal_Proof t rls) prf)),
    (''sizeChangeProc'', Xmlt.many1 ''sizeChangeProc''
      (Xmlt.options [
        (''subtermCriterion'', Xmlt.leaf ''subtermCriterion'' None),
        (''reductionPair'', xml1to2elements ''reductionPair'' 
          (CPF_Parser.ordering_constraint_proof False xml2name) 
          (Xmlt.singleton ''usableRules'' (CPF_Parser.rules xml2name) id) 
          (\<lambda>redp ur. Some (redp,ur)))
      ])
      (CPF_Parser.scg xml2name)
      (\<lambda>version. case version of
        None \<Rightarrow> Size_Change_Subterm_Proc
      | Some redp_ur \<Rightarrow> Size_Change_Redpair_Proc (fst redp_ur) (snd redp_ur))
  )]) id"

abbreviation trs_termination_proof where "trs_termination_proof xml2name 
  xml2trs_termination_proof 
  xml2dp_termination_proof 
  xml2fptrs_termination_proof 
  xml2unknown_proof \<equiv>
  Xmlt.options [(''trsTerminationProof'', Xmlt.singleton ''trsTerminationProof'' (Xmlt.options [
    (''rIsEmpty'', Xmlt.leaf ''rIsEmpty'' R_is_Empty),
    (''semlab'',Xmlt.guard (\<lambda> x. Xml.num_children x = 3)
     (Xmlt.triple ''semlab'' 
       (CPF_Parser.sl_variant xml2name) 
       (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
       xml2trs_termination_proof
       (\<lambda>sli. Semlab sli []))
     (Xmlt.tuple4 ''semlab'' 
       (CPF_Parser.sl_variant xml2name) 
       (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
       (innermostLhss xml2name)
       xml2trs_termination_proof
       (\<lambda>sli lr lq. Semlab sli lq lr))),
    (''split'', Xmlt.triple ''split'' 
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
      xml2trs_termination_proof
      xml2trs_termination_proof
      Split),
    (''dpTrans'', Xmlt.triple ''dpTrans'' 
      (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id)
      (Xmlt.bool ''markedSymbols'')
      xml2dp_termination_proof
      (\<lambda>dps _ prf. DP_Trans default_nfs_dp True dps prf)),
    (''ruleRemoval'', Xmlt.triple ''ruleRemoval''
      (CPF_Parser.ordering_constraint_proof False xml2name) 
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
      xml2trs_termination_proof
      Rule_Removal),
    (''bounds'', Xmlt.change (bounds_info xml2name) Bounds),
    (''stringReversal'', Xmlt.pair ''stringReversal''
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
      xml2trs_termination_proof
      (\<lambda>_. String_Reversal)),
    (''constantToUnary'',Xmlt.tuple4 ''constantToUnary'' 
       (CPF_Parser.plain_var)
       (CPF_Parser.renaming xml2name)
       (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
       xml2trs_termination_proof
       (\<lambda> v ren S p. Constant_String (Const_string_sound_proof v ren S []) p)),
    (''removeNonApplicableRules'', Xmlt.pair ''removeNonApplicableRules''
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
      xml2trs_termination_proof
      Remove_Nonapplicable_Rules),
    (''uncurry'', Xmlt.triple ''uncurry'' 
      (uncurry_info xml2name)
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
      xml2trs_termination_proof
      Uncurry),
    (''flatContextClosure'', Xmlt.triple ''flatContextClosure''
      (CPF_Parser.flat_contexts xml2name)
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
      xml2trs_termination_proof
      Fcc),
    (''switchInnermost'', Xmlt.pair ''switchInnermost''
      (wcr_proof xml2name)
      xml2trs_termination_proof
      Switch_Innermost),
    (''permutingArgumentFilter'', Xmlt.pair ''permutingArgumentFilter''
      (CPF_Parser.afs xml2name)
      xml2trs_termination_proof
      Permuting_AFS),
    (''terminationAssumption'', Xmlt.singleton ''terminationAssumption''
        (xml2inn_trs_assm xml2name)
        (\<lambda> qtrs. Assume_SN qtrs [])),
    (''unknownProof'', Xmlt.many2 ''unknownProof''
        (Xmlt.text ''description'')
        (xml2inn_trs_assm xml2name)
        (sub_proof xml2name
             xml2trs_termination_proof
             xml2dp_termination_proof
             xml2fptrs_termination_proof 
             xml2unknown_proof)
        (\<lambda> desc. Assume_SN))
  ]) id),
  (''relativeTerminationProof'', Xmlt.singleton ''relativeTerminationProof'' (Xmlt.options [
    (''rIsEmpty'', Xmlt.leaf ''rIsEmpty''
      R_is_Empty),
    (''sIsEmpty'', Xmlt.singleton ''sIsEmpty''
      xml2trs_termination_proof
      id),
    (''ruleRemoval'', Xmlt.tuple4 ''ruleRemoval''
      (CPF_Parser.ordering_constraint_proof False xml2name) 
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
      xml2trs_termination_proof
      (\<lambda>rp one two prf. Rule_Removal rp (one @ two) prf)),
    (''semlab'',Xmlt.guard (\<lambda> x. Xml.num_children x = 4) 
      (Xmlt.tuple4 ''semlab'' 
       (CPF_Parser.sl_variant xml2name) 
       (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
       (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
       xml2trs_termination_proof
       (\<lambda>sli lr lrw. Semlab sli [] (lr @ lrw)))
      (Xmlt.tuple5 ''semlab'' 
       (CPF_Parser.sl_variant xml2name) 
       (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
       (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
       (innermostLhss xml2name)
       xml2trs_termination_proof
       (\<lambda>sli lr lrw lq. Semlab sli lq (lr @ lrw)))),
    (''uncurry'', Xmlt.tuple4 ''uncurry'' 
      (uncurry_info xml2name)
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
      xml2trs_termination_proof
      (\<lambda>i r s p. Uncurry i (r @ s) p)),
    (''permutingArgumentFilter'', Xmlt.pair ''permutingArgumentFilter''
      (CPF_Parser.afs xml2name)
      xml2trs_termination_proof
      Permuting_AFS),
    (''flatContextClosure'', Xmlt.tuple4 ''flatContextClosure''
      (CPF_Parser.flat_contexts xml2name)
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
      xml2trs_termination_proof
      (\<lambda> fcc r s. Fcc fcc (r @ s))),
    (''stringReversal'', Xmlt.triple ''stringReversal''
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
      xml2trs_termination_proof
      (\<lambda> _ _. String_Reversal)),
    (''equalityRemoval'',Xmlt.singleton ''equalityRemoval'' 
       xml2trs_termination_proof
       Drop_Equality),
    (''constantToUnary'',Xmlt.tuple5 ''constantToUnary'' 
       (CPF_Parser.plain_var)
       (CPF_Parser.renaming xml2name)
       (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
       (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
       xml2trs_termination_proof
       (\<lambda> v ren S Sw p. Constant_String (Const_string_sound_proof v ren S Sw) p)),
    (''relativeTerminationAssumption'',
      Xmlt.singleton ''relativeTerminationAssumption''
        (xml1to2elements ''trsInput''
          (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
          (Xmlt.singleton ''relativeRules'' (CPF_Parser.rules xml2name) id)
          (\<lambda>r rwo. case rwo of
            None \<Rightarrow> Assume_SN (default_nfs_trs, [], r, []) []
          | Some rw \<Rightarrow> Assume_SN (default_nfs_trs, [], r, rw) [])) id),
    (''unknownProof'', Xmlt.many2 ''unknownProof''
        (Xmlt.text ''description'')
        (xml2inn_trs_assm xml2name)
        (sub_proof xml2name
             xml2trs_termination_proof
             xml2dp_termination_proof
             xml2fptrs_termination_proof 
             xml2unknown_proof)
        (\<lambda> desc. Assume_SN))
  ]) id)]"

abbreviation unknown_proof where "unknown_proof xml2name 
  xml2trs_termination_proof 
  xml2dp_termination_proof 
  xml2fptrs_termination_proof 
  xml2unknown_proof \<equiv>
  Xmlt.singleton ''unknownInputProof'' (Xmlt.options [
    (''unknownAssumption'', Xmlt.singleton ''unknownAssumption''
        xml2unknown_input
        (\<lambda> u. Assume_Unknown u [])),
    (''unknownProof'', Xmlt.many2 ''unknownProof''
        (Xmlt.text ''description'')
        xml2unknown_input
        (sub_proof xml2name
             xml2trs_termination_proof
             xml2dp_termination_proof
             xml2fptrs_termination_proof 
             xml2unknown_proof)
        (\<lambda> desc. Assume_Unknown))
  ]) id"

abbreviation fptrs_termination_proof where "fptrs_termination_proof xml2name 
  xml2trs_termination_proof 
  xml2dp_termination_proof 
  xml2fptrs_termination_proof 
  xml2unknown_proof \<equiv>
  Xmlt.singleton ''trsTerminationProof'' (Xmlt.options [
    (''terminationAssumption'', Xmlt.singleton ''terminationAssumption''
        (xml2fp_trs_assm xml2name)
        (\<lambda> qtrs. Assume_FP_SN qtrs [])),
    (''unknownProof'', Xmlt.many2 ''unknownProof''
        (Xmlt.text ''description'')
        (xml2fp_trs_assm xml2name)
        (sub_proof xml2name
             xml2trs_termination_proof
             xml2dp_termination_proof
             xml2fptrs_termination_proof 
             xml2unknown_proof)
        (\<lambda> desc. Assume_FP_SN))
  ]) id"

(*FIXME: should we set this globally for IsaFoR?*)
(* manually program mutual recursion, as at the moment of writing, it was
   not supported by partial_function *)
declare [[unify_search_bound = 1400, unify_trace_bound = 1400]]

partial_function_mr (sum_bot) 
  xml2dp_termination_proof ::
    "('f :: key, 'l :: key) lab xmlt \<Rightarrow> ('f, 'l, string) dp_termination_proof xmlt" 
  and xml2trs_termination_proof :: 
    "('f :: key, 'l :: key) lab xmlt \<Rightarrow> ('f, 'l, string) trs_termination_proof xmlt"
  and xml2fptrs_termination_proof :: 
    "('f :: key, 'l :: key) lab xmlt \<Rightarrow> ('f, 'l, string) fptrs_termination_proof xmlt" 
  and xml2unknown_proof :: 
    "('f :: key, 'l :: key) lab xmlt \<Rightarrow> ('f, 'l, string) unknown_proof xmlt" where
  "xml2dp_termination_proof xml2name x = dp_termination_proof xml2name 
        (xml2trs_termination_proof xml2name)
        (xml2dp_termination_proof xml2name)  
        (xml2fptrs_termination_proof xml2name)  
        (xml2unknown_proof xml2name) x"
| "xml2trs_termination_proof xml2name x = trs_termination_proof xml2name
        (xml2trs_termination_proof xml2name)
        (xml2dp_termination_proof xml2name)  
        (xml2fptrs_termination_proof xml2name)  
        (xml2unknown_proof xml2name) x"
| "xml2fptrs_termination_proof xml2name x = fptrs_termination_proof xml2name 
        (xml2trs_termination_proof xml2name)
        (xml2dp_termination_proof xml2name)  
        (xml2fptrs_termination_proof xml2name)  
        (xml2unknown_proof xml2name) x"
| "xml2unknown_proof xml2name x = unknown_proof xml2name 
        (xml2trs_termination_proof xml2name)
        (xml2dp_termination_proof xml2name)  
        (xml2fptrs_termination_proof xml2name)  
        (xml2unknown_proof xml2name) x"

text \<open>
  Parse a list of pairs of rules, where the first argument gives the name of the outermost tag
  and the second argument the name of the tag around each pair.
\<close>
definition
  rule_pairs :: "string \<Rightarrow> string \<Rightarrow> 'f xmlt \<Rightarrow> (('f, string) rule \<times> ('f, string) rule) list xmlt"
where
  "rule_pairs s p xml2name =
    Xmlt.many s (Xmlt.pair p
      (CPF_Parser.rule xml2name)
      (CPF_Parser.rule xml2name)
      Pair) id"
hide_const (open) rule_pairs

abbreviation
  sub_cpx_proof ::
    "(xml \<Rightarrow> string +\<^sub>\<bottom> ('f, 'l) lab) \<Rightarrow> (xml \<Rightarrow> string +\<^sub>\<bottom> 'a) \<Rightarrow> xml \<Rightarrow>
    string +\<^sub>\<bottom> ('f, 'l, string, 'a) cpx_assm_proof"
where
  "sub_cpx_proof xml2name xml2complexity_proof x \<equiv>
    Xmlt.pair ''subProof''
      (xml2complexity_input' xml2name)
      xml2complexity_proof
      Complexity_assm_proof
      x"

partial_function (sum_bot) 
  xml2complexity_proof :: "('f,'l)lab xmlt \<Rightarrow> ('f, 'l, string) complexity_proof xmlt"
where
  [code]: "xml2complexity_proof xml2name x = Xmlt.singleton ''complexityProof'' (Xmlt.options [
    (''ruleShifting'', Xmlt.guard (\<lambda> x. Xml.num_children x = 3) 
       (Xmlt.triple ''ruleShifting''
          (CPF_Parser.ordering_constraint_proof True xml2name) 
          (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
          (xml2complexity_proof xml2name)
          (\<lambda> rp del. Rule_Shift_Complexity rp del None))
       (Xmlt.tuple4 ''ruleShifting''
          (CPF_Parser.ordering_constraint_proof True xml2name) 
          (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
          (Xmlt.singleton ''usableRules'' (CPF_Parser.rules xml2name) Some)
          (xml2complexity_proof xml2name)
          Rule_Shift_Complexity)),
    (''usableRules'', Xmlt.pair ''usableRules''
      (Xmlt.singleton ''nonUsableRules'' (CPF_Parser.rules xml2name) id)
      (xml2complexity_proof xml2name)
      Usable_Rules_Complexity),
    (''split'', Xmlt.triple ''split''
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
      (xml2complexity_proof xml2name)
      (xml2complexity_proof xml2name)
      Split_Complexity),
    (''removeNonApplicableRules'', Xmlt.pair ''removeNonApplicableRules''
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
      (xml2complexity_proof xml2name)
      Remove_Nonapplicable_Rules_Complexity),
    (''bounds'', Xmlt.change (bounds_info xml2name) Matchbounds_Complexity), 
    (''relativeBounds'', Xmlt.triple ''relativeBounds''
         (bounds_info xml2name)
         (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
         (xml2complexity_proof xml2name) 
         Matchbounds_Rel_Complexity), 
    (''rIsEmpty'', Xmlt.leaf ''rIsEmpty''
      RisEmpty_Complexity),
    (''dtTransformation'', Xmlt.tuple4 ''dtTransformation''
        (CPF_Parser.rule_pairs ''strictDTs'' ''ruleWithDT'' xml2name)
        (CPF_Parser.rule_pairs ''weakDTs'' ''ruleWithDT'' xml2name)
        (innermostLhss xml2name)
        (xml2complexity_proof xml2name)
        (\<lambda> S W inn p. DT_Transformation (DT_Transformation_Info S W inn) p)),
    (''wdpTransformation'', Xmlt.tuple5 ''wdpTransformation''
      (CPF_Parser.symbols ''compoundSymbols'' xml2name)
      (CPF_Parser.rule_pairs ''strictWDPs'' ''ruleWithWDP'' xml2name)
      (CPF_Parser.rule_pairs ''weakWDPs'' ''ruleWithWDP'' xml2name)
      (innermostLhss xml2name)
      (xml2complexity_proof xml2name)
      (\<lambda>Comp S W Q' p. WDP_Transformation (WDP_Trans_Info (set Comp) S W Q') p)),
    (''unknownProof'', Xmlt.many2 ''unknownProof''
        (Xmlt.text ''description'')
        (xml2complexity_input' xml2name)
        (sub_cpx_proof xml2name
             (xml2complexity_proof xml2name))
        (\<lambda> desc. Complexity_Assumption)),
    (''complexityAssumption'', Xmlt.singleton ''complexityAssumption''
       (xml2complexity_input' xml2name)
       (\<lambda> p. Complexity_Assumption p [] ))
   ]) id x"

definition nonloop :: "'f xmlt \<Rightarrow> ('f,string) non_loop_prf xmlt" where
  "nonloop xml2name =
    Xmlt.tuple6 ''nonLoop''
      (CPF_Parser.pat_rule_prf xml2name)
      (CPF_Parser.subst xml2name)
      (CPF_Parser.subst xml2name)
      (Xmlt.nat ''natural'')
      (Xmlt.nat ''natural'')
      CPF_Parser.pos
      Non_loop_prf"

definition string :: "'f xmlt \<Rightarrow> 'f list xmlt" where
  "string xml2name = Xmlt.many ''string'' xml2name id"

definition oc_srs :: "'f xmlt \<Rightarrow> 'f srs_rule xmlt" where
  "oc_srs xml2name = Xmlt.pair ''overlapClosureSRS'' (string xml2name) (string xml2name) Pair"

definition word_pattern :: "'f xmlt \<Rightarrow> 'f word_pat xmlt" where
  "word_pattern xml2name =
    (let s = string xml2name in
    Xmlt.tuple5 ''wordPattern''
      s
      s
      (Xmlt.nat ''factor'')
      (Xmlt.nat ''constant'')
      s
      (\<lambda> l m f c r. (l,(f,c,m),r)))"

definition derivation_pattern :: "'f xmlt \<Rightarrow> 'f deriv_pat xmlt" where
  "derivation_pattern xml2name = Xmlt.pair ''derivationPattern''
     (word_pattern xml2name) (word_pattern xml2name) Pair"

definition derivation_pattern_proof :: "'f xmlt \<Rightarrow> 'f dp_proof_step xmlt" where
  "derivation_pattern_proof xml2name =
    (let 
      oc = oc_srs xml2name; 
      dp = derivation_pattern xml2name;
      s = string xml2name in
    Xmlt.singleton ''derivationPatternProof'' (Xmlt.options [
    (''OC1'', Xmlt.pair ''OC1'' oc (Xmlt.bool ''isPair'') OC1),
    (''OC2'', Xmlt.tuple6 ''OC2'' oc oc oc s s s OC2),
    (''OC2prime'', Xmlt.tuple6 ''OC2prime'' oc oc oc s s s OC2p),
    (''OC3'', Xmlt.tuple5 ''OC3'' oc oc oc s s  OC3),
    (''OC3prime'', Xmlt.tuple5 ''OC3prime'' oc oc oc s s OC3p),
    (''OCintoDP1'', Xmlt.pair ''OCintoDP1'' dp oc OCDP1),
    (''OCintoDP2'', Xmlt.pair ''OCintoDP2'' dp oc OCDP2),
    (''equivalent'', Xmlt.pair ''equivalent'' dp dp WPEQ),
    (''lift'', Xmlt.pair ''lift'' dp dp Lift),
    (''DP_OC_1_1'', Xmlt.tuple5 ''DP_OC_1_1'' dp dp oc s s DPOC1_1),
    (''DP_OC_1_2'', Xmlt.tuple6 ''DP_OC_1_2'' dp dp oc s s s DPOC1_2),
    (''DP_OC_2'', Xmlt.tuple5 ''DP_OC_2'' dp dp oc s s DPOC2),
    (''DP_OC_3_1'', Xmlt.tuple5 ''DP_OC_3_1'' dp dp oc s s DPOC3_1),
    (''DP_OC_3_2'', Xmlt.tuple6 ''DP_OC_3_2'' dp dp oc s s s DPOC3_2),
    (''DP_DP_1_1'', Xmlt.tuple5 ''DP_DP_1_1'' dp dp dp s s DPDP1_1),
    (''DP_DP_1_2'', Xmlt.tuple5 ''DP_DP_1_2'' dp dp dp s s DPDP1_2),
    (''DP_DP_2_1'', Xmlt.tuple5 ''DP_DP_2_1'' dp dp dp s s DPDP2_1),
    (''DP_DP_2_2'', Xmlt.tuple5 ''DP_DP_2_2'' dp dp dp s s DPDP2_2)
    ]
    ) id)"

definition nonloop_srs_reason :: "'f xmlt \<Rightarrow> ('f dp_proof_step list \<Rightarrow> 'f non_loop_srs_proof)xmlt" where
  "nonloop_srs_reason xml2name =
    (let s = string xml2name in
    Xmlt.options [
      (''selfEmbeddingOC'', Xmlt.triple ''selfEmbeddingOC'' 
        s s s
        (\<lambda> l m r. SE_OC (m, l @ m @ r) l r)),
      (''selfEmbeddingDP'', Xmlt.triple ''selfEmbeddingDP''
        (derivation_pattern xml2name) s s
        SE_DP)
    ])"

definition nonloop_srs :: "'f xmlt \<Rightarrow> 'f non_loop_srs_proof xmlt" where
  "nonloop_srs xml2name = Xmlt.pair ''nonterminatingSRS''
    (Xmlt.many ''derivationPatterns'' (derivation_pattern_proof xml2name) id)
    (nonloop_srs_reason xml2name)
    (\<lambda> list prf. prf list)"

hide_const (open) string
hide_const (open) nonloop_srs

definition
  xml2inn_fp_nt_trs_assm ::
    "('f, 'l) lab xmlt \<Rightarrow> (('f, 'l, string) qtrsLL + ('f, 'l, string) fptrsLL) xmlt"
where
  "xml2inn_fp_nt_trs_assm xml2name x = do {
    inp \<leftarrow> xml2pre_trs_input xml2name x;
    (case inp of 
      Inn_TRS_input (inn, r, None) \<Rightarrow> return (Inl (default_nfs_nt_trs, strategy_to_Q inn r, r))
    | FP_TRS_input (fp, r) \<Rightarrow> return (Inr (strategy_to_fp fp r, r))         
    | _ \<Rightarrow> error ''non-relative TRS expected'')
  }"

definition
  xml2fp_nt_trs_assm :: "('f, 'l) lab xmlt \<Rightarrow> ('f, 'l, string) fptrsLL xmlt"
where
  "xml2fp_nt_trs_assm xml2name x = do {
    inp \<leftarrow> xml2pre_trs_input xml2name x;
    (case inp of 
      FP_TRS_input (fp, r) \<Rightarrow> return (strategy_to_fp fp r, r)         
    | _ \<Rightarrow> error ''outermost/forbidden-pattern/context-sensitive TRS expected'')
  }"

definition
  xml2inn_nt_trs_assm :: "('f,'l)lab xmlt \<Rightarrow> ('f,'l,string)qtrsLL xmlt" where
  "xml2inn_nt_trs_assm xml2name x \<equiv> do {
    inp <- xml2pre_trs_input xml2name x;
    case inp of 
      Inn_TRS_input (inn,r,None) \<Rightarrow> return (default_nfs_nt_trs, strategy_to_Q inn r, r)
    | _ \<Rightarrow> error ''(innermost) TRS without relative rules expected''         
    }"

definition
  xml2inn_rel_nt_trs_assm :: "('f,'l)lab xmlt \<Rightarrow> ('f,'l,string)qreltrsLL xmlt" where
  "xml2inn_rel_nt_trs_assm xml2name x \<equiv> do {
    inp <- xml2pre_trs_input xml2name x;
    case inp of 
      Inn_TRS_input (inn,r,so) \<Rightarrow> return (default_nfs_nt_trs, strategy_to_Q inn r, r, case so of None \<Rightarrow> [] | Some s \<Rightarrow> s)
    | _ \<Rightarrow> error ''(innermost) TRS expected''         
    }"

abbreviation sub_disproof where "sub_disproof xml2name 
  xml2trs_nontermination_proof 
  xml2dp_nontermination_proof
  xml2reltrs_nontermination_proof 
  xml2fp_nontermination_proof
  xml2unknown_disproof
  x \<equiv> 
    let cs = Xml.children x
     in (if length cs \<noteq> 2 then Xmlt.fail ''subProof'' x
      else let inp = take 2 (Xml.tag (cs ! 1)) in
      (if inp = ''tr'' then 
        do {
          io_trs <- xml2inn_fp_nt_trs_assm xml2name (hd cs);
          (case io_trs of
            Inl qtrs \<Rightarrow> Xmlt.change xml2trs_nontermination_proof (\<lambda> prf. Not_SN_assm_proof qtrs prf :: ('f, 'l, string, 'a, 'b, 'c, 'd,'e) generic_assm_proof)
          | Inr fptrs \<Rightarrow> Xmlt.change xml2fp_nontermination_proof (\<lambda> prf. Not_SN_FP_assm_proof fptrs prf)) (cs ! 1)
        }
      else if inp = ''dp'' then
        Xmlt.pair ''subProof'' (xml2dp_input' False xml2name) xml2dp_nontermination_proof (\<lambda> qdp prf. Infinite_assm_proof qdp prf :: ('f, 'l, string, 'a, 'b, 'c, 'd,'e) generic_assm_proof) x
      else if inp = ''re'' then
        Xmlt.pair ''subProof'' (xml2inn_rel_nt_trs_assm xml2name) xml2reltrs_nontermination_proof Not_RelSN_assm_proof x
      else if inp = ''un'' then
        Xmlt.pair ''subProof'' xml2unknown_input xml2unknown_disproof Unknown_assm_proof x
      else Xmlt.fail ''subProof'' x))"

abbreviation dp_nontermination_proof where "dp_nontermination_proof 
  xml2name 
  xml2trs_nontermination_proof
  xml2dp_nontermination_proof 
  xml2reltrs_nontermination_proof
  xml2fp_nontermination_proof
  xml2unknown_disproof
  \<equiv> Xmlt.singleton ''dpNonterminationProof'' (Xmlt.options [
    (''dpRuleRemoval'', Xmlt.guard (\<lambda> x. Xml.num_children x = 3 ) 
       (Xmlt.triple ''dpRuleRemoval''
            (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) Some)
            (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) Some)
            xml2dp_nontermination_proof
            (\<lambda> p r. DP_Rule_Removal (Rule_removal_nonterm_dp_prf p r)))
       (Xmlt.choice ''dpRuleRemoval'' 
         [
          Xmlt.pair ''dpRuleRemoval''
            (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) Some)
            xml2dp_nontermination_proof
            (\<lambda> r. DP_Rule_Removal (Rule_removal_nonterm_dp_prf None r)),
          Xmlt.pair ''dpRuleRemoval''
            (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) Some)
            xml2dp_nontermination_proof
            (\<lambda> p. DP_Rule_Removal (Rule_removal_nonterm_dp_prf p None))
         ])),
    (''loop'', Xmlt.change (CPF_Parser.loop xml2name)
      (\<lambda>(s, rseq, \<sigma>, C). DP_Loop (DP_loop_prf s rseq \<sigma> C))),
    (''nonLoop'', Xmlt.change (CPF_Parser.nonloop xml2name)
      DP_Nonloop),
    (''innermostLhssRemovalProc'', Xmlt.pair ''innermostLhssRemovalProc'' 
      (innermostLhss xml2name)
      xml2dp_nontermination_proof
      (\<lambda> q. DP_Q_Reduction (DP_q_reduction_nonterm_prf q))),
    (''innermostLhssIncreaseProc'', Xmlt.pair ''innermostLhssIncreaseProc'' 
      (innermostLhss xml2name)
      xml2dp_nontermination_proof 
      (\<lambda> q. DP_Q_Increase (Q_increase_nonterm_dp_prf q))),
    (''instantiationProc'', Xmlt.pair ''instantiationProc'' 
      (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id)
      xml2dp_nontermination_proof
      (\<lambda> p. DP_Instantiation (Instantiation_complete_proc_prf p))),
    (''narrowingProc'', Xmlt.tuple4 ''narrowingProc'' 
      (CPF_Parser.rule xml2name)
      CPF_Parser.pos
      (Xmlt.singleton ''narrowings'' (CPF_Parser.rules xml2name) id) 
      xml2dp_nontermination_proof 
      (\<lambda> st po p. DP_Narrowing (Narrowing_complete_proc_prf st po p))),
    (''rewritingProc'', Xmlt.guard (\<lambda> x. Xml.num_children x = 3) 
       (Xmlt.triple ''rewritingProc'' 
         (CPF_Parser.rule xml2name)
         (CPF_Parser.rstep xml2name)
         xml2dp_nontermination_proof
         (\<lambda> (s,t) (p,lr,t'). DP_Rewriting (Rewriting_complete_proc_prf None (s,t) (s,t') (s,t') lr p)))
       (Xmlt.guard (\<lambda> x. Xml.num_children x = 4)
         (Xmlt.guard (\<lambda> x. Xml.tag (Xml.children x ! 2) = ''usableRules'')
           (Xmlt.tuple4 ''rewritingProc'' 
             (CPF_Parser.rule xml2name)
             (CPF_Parser.rstep xml2name)
             (Xmlt.singleton ''usableRules'' (CPF_Parser.rules xml2name) id) 
             xml2dp_nontermination_proof
             (\<lambda> (s,t) (p,lr,t') U. DP_Rewriting (Rewriting_complete_proc_prf (Some U) (s,t) (s,t') (s,t') lr p)))
           (Xmlt.tuple4 ''rewritingProc'' 
             (CPF_Parser.rule xml2name)
             (CPF_Parser.rstep xml2name)
             (CPF_Parser.rule xml2name)
             xml2dp_nontermination_proof 
             (\<lambda> (s,t) (p,lr,t') st''. DP_Rewriting (Rewriting_complete_proc_prf None (s,t) (s,t') st'' lr p))))
       (Xmlt.tuple5 ''rewritingProc'' 
         (CPF_Parser.rule xml2name)
         (CPF_Parser.rstep xml2name)
         (CPF_Parser.rule xml2name)
         (Xmlt.singleton ''usableRules'' (CPF_Parser.rules xml2name) id) 
         xml2dp_nontermination_proof
         (\<lambda> (s,t) (p,lr,t') st'' U. DP_Rewriting (Rewriting_complete_proc_prf (Some U) (s,t) (s,t') st'' lr p))))),
    (''switchFullStrategyProc'', Xmlt.pair ''switchFullStrategyProc'' 
      (wcr_proof xml2name)
      xml2dp_nontermination_proof 
      DP_Termination_Switch),
    (''infinitenessAssumption'', Xmlt.singleton ''infinitenessAssumption''
      (xml2dp_input' False xml2name)
      (\<lambda> qdp. DP_Assume_Infinite qdp [])),
    (''unknownProof'', Xmlt.many2 ''unknownProof''
        (Xmlt.text ''description'')
        (xml2dp_input' False xml2name)
        (sub_disproof xml2name
             xml2trs_nontermination_proof
             xml2dp_nontermination_proof
             xml2reltrs_nontermination_proof
             xml2fp_nontermination_proof
             xml2unknown_disproof)
        (\<lambda> desc. DP_Assume_Infinite)) 
  ]) id"

abbreviation trs_nontermination_proof where "trs_nontermination_proof 
  xml2name 
  xml2trs_nontermination_proof 
  xml2dp_nontermination_proof 
  xml2reltrs_nontermination_proof
  xml2fp_nontermination_proof
  xml2unknown_disproof
  \<equiv> Xmlt.singleton ''trsNonterminationProof'' (Xmlt.options [
  (''variableConditionViolated'', Xmlt.leaf ''variableConditionViolated''
    TRS_Not_Well_Formed),
  (''loop'', Xmlt.change (CPF_Parser.loop xml2name)
    (\<lambda>(s, rseq, \<sigma>, C). TRS_Loop (TRS_loop_prf s (map (\<lambda>(x, y, _, z). (x, y, z)) rseq) \<sigma> C))),
  (''nonLoop'', Xmlt.change (CPF_Parser.nonloop xml2name)
    TRS_Nonloop),
  (''nonterminatingSRS'', Xmlt.change (CPF_Parser.nonloop_srs xml2name)
    TRS_Nonloop_SRS),
  (''ruleRemoval'', Xmlt.pair ''ruleRemoval''
    (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
    xml2trs_nontermination_proof
    (\<lambda> r. TRS_Rule_Removal (Rule_removal_nonterm_trs_prf r))),
  (''dpTrans'', Xmlt.triple ''dpTrans''
    (Xmlt.singleton ''dps'' (CPF_Parser.rules xml2name) id)
    (Xmlt.bool ''markedSymbols'')
    xml2dp_nontermination_proof
    (\<lambda>p _. TRS_DP_Trans (DP_trans_nontermination_tt_prf p))),
  (''stringReversal'', Xmlt.pair ''stringReversal''
    (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
    xml2trs_nontermination_proof
    (\<lambda>_. TRS_String_Reversal)),
  (''constantToUnary'',Xmlt.tuple4 ''constantToUnary'' 
     (CPF_Parser.plain_var)
     (CPF_Parser.renaming xml2name)
     (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
     xml2trs_nontermination_proof
     (\<lambda> v ren S p. TRS_Constant_String (Const_string_complete_proof v ren S) p)),
  (''innermostLhssIncrease'', Xmlt.pair ''innermostLhssIncrease'' 
    (innermostLhss xml2name)
    xml2trs_nontermination_proof
    (\<lambda> q. TRS_Q_Increase (Q_increase_nonterm_trs_prf q))),
  (''switchFullStrategy'', Xmlt.pair ''switchFullStrategy'' 
    (wcr_proof xml2name)
    xml2trs_nontermination_proof 
    TRS_Termination_Switch),
  (''uncurry'',  Xmlt.triple ''uncurry'' 
      (uncurry_info xml2name) 
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id) 
      xml2trs_nontermination_proof
      (\<lambda> i r. TRS_Uncurry (Uncurry_nt_proof i r))),
  (''nonterminationAssumption'', Xmlt.singleton ''nonterminationAssumption''
    (xml2inn_nt_trs_assm xml2name)
    (\<lambda> qtrs. TRS_Assume_Not_SN qtrs [])),
  (''unknownProof'', Xmlt.many2 ''unknownProof''
    (Xmlt.text ''description'')
    (xml2inn_nt_trs_assm xml2name)
    (sub_disproof xml2name
         xml2trs_nontermination_proof
         xml2dp_nontermination_proof
         xml2reltrs_nontermination_proof
         xml2fp_nontermination_proof
         xml2unknown_disproof)
    (\<lambda> desc. TRS_Assume_Not_SN)) 
  ]) id"

abbreviation reltrs_nontermination_proof where "reltrs_nontermination_proof 
  xml2name 
  xml2trs_nontermination_proof 
  xml2dp_nontermination_proof 
  xml2reltrs_nontermination_proof
  xml2fp_nontermination_proof
  xml2unknown_disproof
  \<equiv> Xmlt.singleton ''relativeNonterminationProof'' (Xmlt.options [
    (''variableConditionViolated'', Xmlt.leaf ''variableConditionViolated''
      Rel_Not_Well_Formed),
    (''loop'', Xmlt.change (CPF_Parser.loop xml2name)
      (\<lambda>(s, rseq, \<sigma>, C). Rel_Loop (Rel_trs_loop_prf s rseq \<sigma> C))),
    (''trsNonterminationProof'', Xmlt.change xml2trs_nontermination_proof Rel_R_Not_SN),
    (''ruleRemoval'', Xmlt.triple ''ruleRemoval''
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) Some)
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) Some)
      xml2reltrs_nontermination_proof
      (\<lambda> r s. Rel_Rule_Removal (Rule_removal_nonterm_reltrs_prf r s))),
    (''stringReversal'', Xmlt.triple ''stringReversal''
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) Some)
      (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) Some)
      xml2reltrs_nontermination_proof
      (\<lambda> _ _. Rel_TRS_String_Reversal)),
    (''nonterminationAssumption'', Xmlt.singleton ''nonterminationAssumption''
      (xml2inn_rel_nt_trs_assm xml2name)
      (\<lambda> qtrs. Rel_TRS_Assume_Not_SN qtrs [])),
    (''unknownProof'', Xmlt.many2 ''unknownProof''
      (Xmlt.text ''description'')
      (xml2inn_rel_nt_trs_assm xml2name)
      (sub_disproof xml2name
           xml2trs_nontermination_proof
           xml2dp_nontermination_proof
           xml2reltrs_nontermination_proof
           xml2fp_nontermination_proof
           xml2unknown_disproof)
      (\<lambda> desc. Rel_TRS_Assume_Not_SN))]) id"

abbreviation fp_nontermination_proof where "fp_nontermination_proof 
  xml2name 
  xml2trs_nontermination_proof 
  xml2dp_nontermination_proof 
  xml2reltrs_nontermination_proof
  xml2fp_nontermination_proof
  xml2unknown_disproof
  \<equiv> Xmlt.singleton ''trsNonterminationProof'' (Xmlt.options [
  (''loop'', Xmlt.change (CPF_Parser.loop xml2name)
    (\<lambda>(s, rseq, \<sigma>, C). FPTRS_Loop (FP_loop_prf C \<sigma> s (map (\<lambda>(x, y, _, z). (x, y, z)) rseq)))),
  (''ruleRemoval'', Xmlt.pair ''ruleRemoval''
    (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
    xml2fp_nontermination_proof
    (\<lambda> r. FPTRS_Rule_Removal (Rule_removal_nonterm_trs_prf r))),
  (''nonterminationAssumption'', Xmlt.singleton ''nonterminationAssumption''
      (xml2fp_nt_trs_assm xml2name)
      (\<lambda> qtrs. FPTRS_Assume_Not_SN qtrs [])),
  (''unknownProof'', Xmlt.many2 ''unknownProof''
      (Xmlt.text ''description'')
      (xml2fp_nt_trs_assm xml2name)
      (sub_disproof xml2name
           xml2trs_nontermination_proof
           xml2dp_nontermination_proof
           xml2reltrs_nontermination_proof
           xml2fp_nontermination_proof
           xml2unknown_disproof)
      (\<lambda> desc. FPTRS_Assume_Not_SN)) 
  ]) id"

abbreviation unknown_disproof where "unknown_disproof xml2name 
  xml2trs_nontermination_proof 
  xml2dp_nontermination_proof 
  xml2reltrs_nontermination_proof
  xml2fp_nontermination_proof
  xml2unknown_disproof \<equiv>
  Xmlt.singleton ''unknownInputProof'' (Xmlt.options [
    (''unknownAssumption'', Xmlt.singleton ''unknownAssumption''
        xml2unknown_input
        (\<lambda> u. Assume_NT_Unknown u [])),
    (''unknownProof'', Xmlt.many2 ''unknownProof''
        (Xmlt.text ''description'')
        xml2unknown_input
        (sub_disproof xml2name
             xml2trs_nontermination_proof
             xml2dp_nontermination_proof
             xml2reltrs_nontermination_proof
             xml2fp_nontermination_proof
             xml2unknown_disproof)
        (\<lambda> desc. Assume_NT_Unknown))
  ]) id"

partial_function_mr (sum_bot)
  xml2dp_nontermination_proof ::
    "('f, 'l) lab xmlt \<Rightarrow> ('f, 'l, string) dp_nontermination_proof xmlt" and 
  xml2trs_nontermination_proof :: 
    "('f, 'l) lab xmlt \<Rightarrow> ('f, 'l, string) trs_nontermination_proof xmlt" and
  xml2reltrs_nontermination_proof :: 
    "('f, 'l) lab xmlt \<Rightarrow> ('f, 'l, string) reltrs_nontermination_proof xmlt" and
  xml2fp_nontermination_proof :: 
    "('f,'l)lab xmlt \<Rightarrow> ('f, 'l, string) fp_nontermination_proof xmlt" and
  xml2unknown_disproof :: "('f, 'l) lab xmlt \<Rightarrow> ('f, 'l, string) neg_unknown_proof xmlt" 
  where
  "xml2dp_nontermination_proof xml2name x = dp_nontermination_proof xml2name 
     (xml2trs_nontermination_proof xml2name)
     (xml2dp_nontermination_proof xml2name)
     (xml2reltrs_nontermination_proof xml2name)
     (xml2fp_nontermination_proof xml2name)
     (xml2unknown_disproof xml2name) x"
| "xml2trs_nontermination_proof xml2name x = trs_nontermination_proof xml2name 
     (xml2trs_nontermination_proof xml2name)
     (xml2dp_nontermination_proof xml2name)
     (xml2reltrs_nontermination_proof xml2name)
     (xml2fp_nontermination_proof xml2name)
     (xml2unknown_disproof xml2name) x"
| "xml2reltrs_nontermination_proof xml2name x = reltrs_nontermination_proof xml2name 
     (xml2trs_nontermination_proof xml2name)
     (xml2dp_nontermination_proof xml2name)
     (xml2reltrs_nontermination_proof xml2name)
     (xml2fp_nontermination_proof xml2name)
     (xml2unknown_disproof xml2name) x"
| "xml2fp_nontermination_proof xml2name x = fp_nontermination_proof xml2name 
     (xml2trs_nontermination_proof xml2name)
     (xml2dp_nontermination_proof xml2name)
     (xml2reltrs_nontermination_proof xml2name)
     (xml2fp_nontermination_proof xml2name)
     (xml2unknown_disproof xml2name) x"
| "xml2unknown_disproof xml2name x = unknown_disproof xml2name 
     (xml2trs_nontermination_proof xml2name)
     (xml2dp_nontermination_proof xml2name)
     (xml2reltrs_nontermination_proof xml2name)
     (xml2fp_nontermination_proof xml2name)
     (xml2unknown_disproof xml2name) x"

definition rule_labeling_function :: "'a xmlt \<Rightarrow> ('a, string) rule_lab_repr xmlt"
where
  "rule_labeling_function xml2name \<equiv> Xmlt.many ''ruleLabelingFunction''
   (Xmlt.pair ''ruleLabelingFunctionEntry''
    (CPF_Parser.rule xml2name)
    (Xmlt.nat ''label'')
    Pair) id"
hide_const (open) rule_labeling_function

partial_function (sum_bot)
  xml2cr_proof :: "('f :: key,label_type)lab xmlt \<Rightarrow> ('f, label_type,string) cr_proof xmlt"
where
  [code]: "xml2cr_proof xml2name x = Xmlt.singleton ''crProof'' (Xmlt.options [
    (''wcrAndSN'', Xmlt.pair ''wcrAndSN''
         (wcr_proof xml2name)
         (xml2trs_termination_proof xml2name)
      SN_WCR),
    (''orthogonal'', Xmlt.leaf ''orthogonal'' Weakly_Orthogonal),
    (''stronglyClosed'', Xmlt.change (Xmlt.nat ''stronglyClosed'') Strongly_Closed),
    (''ruleLabeling'', Xmlt.pair ''ruleLabeling''
        (CPF_Parser.rule_labeling_function xml2name)
        (joinable_critical_pairs xml2name)
      (\<lambda> rl js. Rule_Labeling rl js None)),
    (''decreasingDiagrams'', Xmlt.guard (\<lambda>x. Xml.num_children x = 2)
      (Xmlt.pair ''decreasingDiagrams''
        (xml2trs_termination_proof xml2name)
        (Xmlt.pair ''ruleLabeling'' 
          (CPF_Parser.rule_labeling_function xml2name) 
          (joinable_critical_pairs xml2name)
        (\<lambda> rl js. Rule_Labeling rl js))
      (\<lambda> prf rljs. rljs (Some prf)))
      (Xmlt.singleton ''decreasingDiagrams''
        (Xmlt.pair ''ruleLabeling'' 
          (CPF_Parser.rule_labeling_function xml2name) 
          (joinable_critical_pairs xml2name)
        (\<lambda> rl js. Rule_Labeling rl js None))
        id)),
    (''redundantRules'', Xmlt.xml3to4elements ''redundantRules''
        (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
        (Xmlt.nat ''nrSteps'')
        (Xmlt.many ''conversions'' (CPF_Parser.conversion xml2name) id)
        (xml2cr_proof xml2name)
      (\<lambda> trs n convs prf. case convs of
          None \<Rightarrow> Redundant_Rules trs n [] prf
        | Some cs \<Rightarrow> Redundant_Rules trs n cs prf))
  ]) id x"

definition compatible_ta :: "string \<Rightarrow> 'f xmlt \<Rightarrow> ((string,'f)tree_automaton \<times> string ta_relation) xmlt" where
  "compatible_ta tag xml2name \<equiv> Xmlt.guard (\<lambda> x. Xml.num_children x = 1) 
    (Xmlt.singleton tag
      (tree_automaton (ta_normal_lhs xml2name))
      (\<lambda> ta. (ta, Id_Relation)))
    (Xmlt.pair tag
      (tree_automaton (ta_normal_lhs xml2name))
      closed_criterion  
      Pair)
    "

partial_function (sum_bot) xml2non_join_info :: "string lt xmlt \<Rightarrow> (string lt,string,string) non_join_info xmlt" where
  [code]: "xml2non_join_info xml2name x = (Xmlt.options [
              (''distinctNormalForms'', Xmlt.leaf ''distinctNormalForms'' Diff_NFs),
              (''capNotUnif'', Xmlt.leaf ''capNotUnif'' Tcap_Not_Unif),
              (''subterm'', Xmlt.pair ''subterm'' 
                  CPF_Parser.pos
                  (xml2non_join_info xml2name)
                  Subterm_NJ),
              (''grounding'', Xmlt.pair ''grounding'' 
                  (CPF_Parser.subst xml2name)
                  (xml2non_join_info xml2name)
                  Grounding),
              (''emptyTreeAutomataIntersection'', Xmlt.pair ''emptyTreeAutomataIntersection''
                  (compatible_ta ''firstAutomaton'' xml2name)
                  (compatible_ta ''secondAutomaton'' xml2name)
                  (\<lambda> (ta1,rel1) (ta2,rel2). Tree_Aut_Intersect_Empty ta1 rel1 ta2 rel2)),
              (''differentInterpretation'', Xmlt.singleton ''differentInterpretation''
                  (CPF_Parser.sl_variant xml2name)
                  Finite_Model_Gt),
              (''strictDecrease'', Xmlt.singleton ''strictDecrease''
                  (CPF_Parser.ordering_constraint_proof False xml2name)
                  Reduction_Pair_Gt),
              (''argumentFilterNonJoin'', Xmlt.pair ''argumentFilterNonJoin''
                  (CPF_Parser.afs xml2name)
                  (xml2non_join_info xml2name)
                  Argument_Filter_NJ),
              (''usableRulesNonJoin'', Xmlt.guard (\<lambda> x. Xml.num_children x = 1) 
                  (Xmlt.singleton ''usableRulesNonJoin''
                    (xml2non_join_info xml2name)
                    Usable_Rules_Reach_NJ)
                  (Xmlt.triple ''usableRulesNonJoin''
                    (Xmlt.options 
                      [(''left'', Xmlt.leaf ''left'' True),
                       (''right'', Xmlt.leaf ''right'' False)])
                    (Xmlt.singleton ''usableRules'' (CPF_Parser.rules xml2name) id)
                    (xml2non_join_info xml2name)
                    (\<lambda> left U p. Usable_Rules_Reach_Unif_NJ (if left then Inl U else Inr U) p)))
           ]) x"

partial_function (sum_bot)
  xml2ncr_proof :: "(string,label_type)lab xmlt \<Rightarrow> (string, label_type,string,string) ncr_proof xmlt"
where
  [code]: "xml2ncr_proof xml2name x = (let rew = CPF_Parser.rsteps xml2name in
    Xmlt.singleton ''crDisproof'' (Xmlt.options [
    (''nonWcrAndSN'', Xmlt.pair ''nonWcrAndSN''
         (\<lambda> _. return ())
         (xml2trs_termination_proof xml2name)
      (\<lambda> _. SN_NWCR)),
    (''nonJoinableFork'', Xmlt.triple ''nonJoinableFork'' rew rew 
           (xml2non_join_info xml2name)
           (\<lambda> (s,seq1) (_,seq2). Non_Join s seq1 seq2)),
    (''modularityDisjoint'', Xmlt.pair ''modularityDisjoint''
        (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
        (xml2ncr_proof xml2name)
        NCR_Disj_Subtrs),
    (''redundantRules'', Xmlt.triple ''redundantRules''
        (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
        (Xmlt.nat ''nrSteps'')
        (xml2ncr_proof xml2name)
      NCR_Redundant_Rules)
  ]) id x)"


definition
  xml2completion_proof :: "('f :: key,label_type)lab xmlt \<Rightarrow> ('f, label_type,string) completion_proof xmlt"
where
  "xml2completion_proof xml2name = Xmlt.triple ''completionProof'' 
     (wcr_proof xml2name)
     (xml2trs_termination_proof xml2name)
     (xml1to2elements ''equivalenceProof''
       (CPF_Parser.subsumption_proof xml2name)
       (CPF_Parser.subsumption_proof xml2name)
       Pair)
     (\<lambda> w t (s1,s2). SN_WCR_Eq w t s1 s2)"

definition "unraveling_info xml2name =
  (Xmlt.many ''unravelingInformation'' (Xmlt.many1 ''unravelingEntry''
           (CPF_Parser.crule xml2name)
           (CPF_Parser.rule xml2name)
           Pair) id)"

hide_const(open) unraveling_info

definition
  xml2quasi_reductive_proof :: "('f :: key,label_type)lab xmlt \<Rightarrow> ('f, label_type,string) quasi_reductive_proof xmlt"
where
  "xml2quasi_reductive_proof xml2name \<equiv> Xmlt.singleton ''quasiReductiveProof'' (Xmlt.options [
    (''unraveling'', Xmlt.pair ''unraveling'' 
        (CPF_Parser.unraveling_info xml2name)        
        (xml2trs_termination_proof xml2name)
        Unravel)
  ]) id"

definition xml2non_reachability_proof :: "non_reachability_proof xmlt"
where
  "xml2non_reachability_proof = Xmlt.singleton ''nonReachabilityProof'' (Xmlt.options [
    (''notReachableTcap'', Xmlt.leaf ''notReachableTcap'' Not_Reachable_Tcap)
  ]) id"

definition
  xml2infeasibility_proof ::
    "('f :: key, label_type) lab xmlt \<Rightarrow> (('f, label_type) lab, string) infeasibility_proof xmlt"
where
  "xml2infeasibility_proof xml2name = Xmlt.singleton ''infeasibilityProof'' (Xmlt.options [
    (''compoundConditionsNotReachable'', Xmlt.pair ''compoundConditionsNotReachable''
     xml2name
     xml2non_reachability_proof
     Compound_Conditions_Not_Reachable)
  ]) id"

definition
  xml2infeasible_conds ::
    "('f :: key, label_type) lab xmlt \<Rightarrow>
      ((('f, label_type) lab, string) rules \<times>
      (('f, label_type) lab, string) infeasibility_proof) list xmlt"
where
  "xml2infeasible_conds xml2name = Xmlt.many ''infeasibleConditions''
    (Xmlt.pair ''infeasibleCondition''
      (CPF_Parser.rules xml2name)
      (xml2infeasibility_proof xml2name)
      Pair)
    id"

definition
  xml2conditional_cr_proof ::
    "('f :: key, label_type) lab xmlt \<Rightarrow> ('f, label_type, string) conditional_cr_proof xmlt"
where
  "xml2conditional_cr_proof xml2name \<equiv> Xmlt.singleton ''conditionalCrProof'' (Xmlt.options [
    (''unraveling'', Xmlt.pair ''unraveling''
        (CPF_Parser.unraveling_info xml2name)
        (xml2cr_proof xml2name)
        Unravel_CR),
    (''almostOrthogonal'', Xmlt.leaf ''almostOrthogonal'' Almost_Orthogonal_CR),
    (''almostOrthogonalModuloInfeasibility'',
      Xmlt.singleton ''almostOrthogonalModuloInfeasibility''
        (xml2infeasible_conds xml2name) Almost_Orthogonal_Modulo_Infeasibility_CR)
  ]) id"

partial_function (sum_bot) xml2eq_proof :: "'a xmlt \<Rightarrow> ('a, string) eq_proof xmlt" where
  [code]: "xml2eq_proof xml2name x = Xmlt.options [
    (''refl'', Xmlt.singleton ''refl'' (CPF_Parser.term xml2name) Equational_Reasoning.Refl),
    (''sym'', Xmlt.singleton ''sym'' (xml2eq_proof xml2name) Sym),
    (''trans'', Xmlt.pair ''trans'' (xml2eq_proof xml2name) (xml2eq_proof xml2name) Trans),
    (''assm'', Xmlt.pair ''assm'' (CPF_Parser.rule xml2name) (CPF_Parser.subst xml2name) (\<lambda>r s. Assm r (mk_subst Var s))),
    (''cong'', Xmlt.many1 ''cong'' (xml2name) (xml2eq_proof xml2name) Cong)
  ] x"
 
definition
  xml2equational_proof :: "('f :: key, label_type) lab xmlt \<Rightarrow> ('f, label_type, string) equational_proof xmlt"
where
  "xml2equational_proof xml2name =
    Xmlt.singleton ''equationalProof'' (Xmlt.options [
      (''equationalProofTree'', Xmlt.singleton ''equationalProofTree'' (xml2eq_proof xml2name) Equational_Proof_Tree),
      (''conversion'', Xmlt.change (CPF_Parser.conversion xml2name) Conversion),
      (''subsumptionProof'', Xmlt.change (CPF_Parser.subsumption_proof xml2name) Conversion_With_History),
      (''completionAndNormalization'', Xmlt.pair ''completionAndNormalization''
        (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
        (xml2completion_proof xml2name)
        Completion_and_Normalization)
    ]) id"

definition
  xml2equational_disproof :: "('f :: key, label_type) lab xmlt \<Rightarrow> ('f, label_type, string) equational_disproof xmlt"
where
  "xml2equational_disproof xml2name =
    Xmlt.singleton ''equationalDisproof'' (Xmlt.options [
      (''completionAndNormalization'', Xmlt.pair ''completionAndNormalization''
        (Xmlt.singleton ''trs'' (CPF_Parser.rules xml2name) id)
        (xml2completion_proof xml2name)
        Completion_and_Normalization_Different)
    ]) id"

type_synonym lsymbol = "(string, label_type) lab"

fun
  xml2cert_problem :: "string xmlt \<Rightarrow> lsymbol xmlt \<Rightarrow> (string, label_type, string) cert_problem xmlt"
where
  "xml2cert_problem xml2uname xml2name (XML name egal cs) =
  (if name = ''certificationProblem'' \<and> length cs \<ge> 3 then do {
    inp     \<leftarrow> xml2input xml2name (cs ! 0);
    version \<leftarrow> Xmlt.text ''cpfVersion'' (cs!1);
    case inp of
      Inn_TRS_input (q,r,s_opt) \<Rightarrow>
      (case s_opt of 
        None \<Rightarrow> Xmlt.singleton ''proof'' (
        Xmlt.options [
          (''trsTerminationProof'',
            Xmlt.change (xml2trs_termination_proof xml2name) (\<lambda>p. TRS_Termination_Proof default_nfs_trs q r None p)),
          (''crProof'',
            if strategy_to_Q q r = []
              then Xmlt.change (xml2cr_proof xml2name) (\<lambda>p. TRS_Confluence_Proof False r p)
              else (\<lambda> _. error ''strategies for confluence unsupported'')),
          (''crDisproof'',
            if strategy_to_Q q r = []
              then Xmlt.change (xml2ncr_proof xml2name) (\<lambda>p. TRS_Non_Confluence_Proof False r p)
              else (\<lambda> _. error ''strategies for confluence unsupported'')),
          (''trsNonterminationProof'',
            Xmlt.change (xml2trs_nontermination_proof xml2name) (\<lambda>p. TRS_Nontermination_Proof default_nfs_trs q r p))]
      ) id (cs!2)
      | Some s \<Rightarrow> Xmlt.singleton ''proof'' (
        Xmlt.options [
          (''relativeTerminationProof'',
            Xmlt.change (xml2trs_termination_proof xml2name) (\<lambda>p. TRS_Termination_Proof default_nfs_trs q r (Some s) p)),
          (''relativeNonterminationProof'', 
            Xmlt.change (xml2reltrs_nontermination_proof xml2name) (\<lambda>p. Relative_TRS_Nontermination_Proof default_nfs_trs q r s p))]
      ) id (cs!2))
    | DP_input (m,p,q,r) \<Rightarrow> Xmlt.singleton ''proof'' (
      Xmlt.options [
        (''dpProof'',
          Xmlt.change (xml2dp_termination_proof xml2name) (\<lambda>prf. DP_Termination_Proof default_nfs_dp m p [] q [] r prf)),
        (''dpNonterminationProof'',
          Xmlt.change (xml2dp_nontermination_proof xml2name) (\<lambda>prf. DP_Nontermination_Proof default_nfs_nt_dp m p q r prf))]
    ) id (cs!2)
    | COMP_input (e, r) \<Rightarrow> Xmlt.singleton ''proof'' (
      Xmlt.options [
        (''completionProof'',
          Xmlt.change (xml2completion_proof xml2name) (\<lambda>prf. Completion_Proof e r prf))]
    ) id (cs!2)
    | EQ_input (eqs, eq) \<Rightarrow> Xmlt.singleton ''proof'' (
      Xmlt.options [
        (''equationalProof'',
          Xmlt.change (xml2equational_proof xml2name) (\<lambda>prf. Equational_Proof eqs eq prf)),
        (''equationalDisproof'',
          Xmlt.change (xml2equational_disproof xml2name) (\<lambda>prf. Equational_Disproof eqs eq prf))
        ]
    ) id (cs!2)
    | CPX_input (q,r,s, cm, cc) \<Rightarrow> Xmlt.singleton ''proof'' (
          Xmlt.change (xml2complexity_proof xml2name) (\<lambda>prf. Complexity_Proof q r s cm cc prf)
    ) id (cs!2)
    | FP_TRS_input (Outermost,r) \<Rightarrow> Xmlt.singleton ''proof'' (
          Xmlt.options [
          (''trsTerminationProof'',
            Xmlt.change (xml2fptrs_termination_proof xml2name) (\<lambda>prf. Outermost_Termination_Proof r prf)),
          (''trsNonterminationProof'',
            Xmlt.change (xml2fp_nontermination_proof xml2name) (\<lambda>prf. Outermost_Nontermination_Proof r prf))]
    ) id (cs!2)
    | FP_TRS_input (Context_Sensitive \<mu>,r) \<Rightarrow> Xmlt.singleton ''proof'' (
          Xmlt.options [
          (''trsTerminationProof'',
            Xmlt.change (xml2fptrs_termination_proof xml2name) (\<lambda>prf. CS_Termination_Proof \<mu> r prf)),
          (''trsNonterminationProof'',
            Xmlt.change (xml2fp_nontermination_proof xml2name) (\<lambda>prf. CS_Nontermination_Proof \<mu> r prf))]
    ) id (cs!2)
    | FP_TRS_input (Forbidden_Patterns p,r) \<Rightarrow> Xmlt.singleton ''proof'' (
          Xmlt.options [
          (''trsTerminationProof'',
            Xmlt.change (xml2fptrs_termination_proof xml2name) (\<lambda>prf. FP_Termination_Proof p r prf)),
          (''trsNonterminationProof'',
            Xmlt.change (xml2fp_nontermination_proof xml2name) (\<lambda>prf. FP_Nontermination_Proof p r prf))]
    ) id (cs!2)
    | CTRS_input ctrs \<Rightarrow> Xmlt.singleton ''proof'' (
          Xmlt.options [
            (''quasiReductiveProof'', 
              Xmlt.change (xml2quasi_reductive_proof xml2name) (\<lambda>prf. Quasi_Reductive_Proof ctrs prf)),
            (''conditionalCrProof'', 
              Xmlt.change (xml2conditional_cr_proof xml2name) (\<lambda>prf. Conditional_CR_Proof ctrs prf))
          ]
    ) id (cs!2)
    | TA_input ta r \<Rightarrow> Xmlt.singleton ''proof'' (
          Xmlt.singleton ''treeAutomatonClosedProof''
            (Xmlt.bool ''decisionProcedure'') (Tree_Automata_Closed_Proof ta r)
          ) id (cs ! 2)
    | Unknown_input u \<Rightarrow> Xmlt.singleton ''proof'' (
          Xmlt.choice ''unknownInput'' [
            Xmlt.change (xml2unknown_proof xml2name) (\<lambda>prf. Unknown_Proof u prf),
            Xmlt.change (xml2unknown_disproof xml2name) (\<lambda>prf. Unknown_Disproof u prf)
          ]
    ) id (cs!2)
  } else error (''could not parse root element <'' @ name @ ''>''))"
| "xml2cert_problem _ _ xml = Xmlt.fail ''certificationProblem'' xml"

partial_function (sum_bot)
  plain_name :: "lsymbol xmlt"
where
  [code]: "plain_name x = Xmlt.options [
     (''name'', Xmlt.change (Xmlt.text ''name'') UnLab),
     (''sharp'', Xmlt.singleton ''sharp'' plain_name Sharp),
     (''labeledSymbol'', Xmlt.pair ''labeledSymbol'' plain_name (
           Xmlt.options [
             (''numberLabel'', Xmlt.many ''numberLabel'' (Xmlt.nat ''number'') Inl),
             (''symbolLabel'', Xmlt.many ''symbolLabel'' plain_name Inr)
                       ])
     (\<lambda>f l. case l of
       Inl ls  \<Rightarrow> Lab f ls
     | Inr ls \<Rightarrow> FunLab f ls))
   ] x"

hide_const (open) plain_name

primrec
  xmldoc2cert_problem ::
    "xmldoc \<Rightarrow> string +\<^sub>\<bottom> (string, label_type, string) cert_problem"
where
  "xmldoc2cert_problem (XMLDOC header xml) = debug
    ''0''
    ''parsing xml to certification problem''
    (xml2cert_problem (Xmlt.text ''name'') CPF_Parser.plain_name xml)"

(*include debug information for xml-parsing*)
definition doc_of_string :: "string \<Rightarrow> string + xmldoc" where
  "doc_of_string = debug ''0'' ''parsing txt to xml'' Xml.doc_of_string"

definition
  parse_cert_problem ::
    "string \<Rightarrow> string +\<^sub>\<bottom> (string, label_type, string) cert_problem"
where
  "parse_cert_problem s =
    (case doc_of_string s of
      Inl x \<Rightarrow> error x
    | Inr y \<Rightarrow> xmldoc2cert_problem y)"

(* ... to xml *)
primrec xml_lab :: "('f :: show, ('l :: show)list)lab \<Rightarrow> xml" where
  "xml_lab (UnLab x) = XML ''name'' [] [XML_text (show x)]"
| "xml_lab (Sharp x) = XML ''sharp'' [] [xml_lab x]"
| "xml_lab (FunLab x l) = XML ''labeledSymbol'' [] [xml_lab x, XML ''symbolLabel'' [] (map xml_lab l)]"
| "xml_lab (Lab x l) = XML ''labeledSymbol'' [] [xml_lab x, XML ''numberLabel'' [] (map (\<lambda> n. XML ''number'' [] [XML_text (show n)]) l)]"

definition xml_signature :: "(('f :: show,('l :: show)list)lab \<times> nat)list \<Rightarrow> xml" where
  "\<And>fs. xml_signature fs \<equiv> XML ''signature'' [] (map (\<lambda> (f,n). XML ''symbol'' [] [xml_lab f, XML ''arity'' [] [XML_text (shows n Nil)]]) fs)"

primrec xml_complexity_measure :: "(('f :: show,('l :: show)list)lab,'v)complexity_measure \<Rightarrow> xml" where
  "xml_complexity_measure (Derivational_Complexity F) = XML ''derivationalComplexity'' [] [xml_signature F]"
 | "xml_complexity_measure (Runtime_Complexity C D) = XML ''runtimeComplexity'' [] [xml_signature C, xml_signature D]"

primrec xml_complexity_class :: "complexity_class \<Rightarrow> xml" where
  "xml_complexity_class (Comp_Poly n) = XML ''polynomial'' [] [XML_text (shows n Nil)]"

definition xml_single_pos :: "nat \<Rightarrow> xml" where
  "xml_single_pos i \<equiv> XML ''position'' [] [XML_text (show (Suc i))]"

definition xml_pos :: "pos \<Rightarrow> xml" where
  "xml_pos p \<equiv>  XML ''positionInTerm'' [] (map xml_single_pos (to_list p))"

primrec xml_term :: "(('f :: show,('l :: show)list)lab, 'v :: show)term \<Rightarrow> xml" where
  "xml_term (Var x) = XML ''var'' [] [XML_text (show x)]" 
| "xml_term (Fun f ts) = XML ''funapp'' [] 
    (xml_lab f # 
      (map (\<lambda> t. XML ''arg'' [] [xml_term t]) ts))"

definition xml_rule :: "(('f :: show,('l :: show)list)lab,'c :: show) rule \<Rightarrow> xml" where
  "xml_rule \<equiv> \<lambda> (l, r). XML ''rule'' [] [
    XML ''lhs'' [] [xml_term l] ,         
    XML ''rhs'' [] [xml_term r] ]"

definition xml_rules :: "string \<Rightarrow> (('f :: show,('l :: show)list)lab,'c :: show) rules \<Rightarrow> xml" where
  "xml_rules tag rls = XML tag Nil [XML ''rules'' [] (map xml_rule rls)]"

definition xml_crule :: "(('f :: show,('l :: show)list)lab,'c :: show) crule \<Rightarrow> xml" where
  "xml_crule \<equiv> \<lambda> (c, rs). XML ''conditionalRule'' [] (map xml_rule (c # rs))"

definition xml_forbidden_pattern :: "(('f :: show,('l :: show) list)lab,'v :: show) forb_pattern \<Rightarrow> xml" where
  "xml_forbidden_pattern \<equiv> \<lambda> (C,t,l). XML ''forbiddenPattern'' [] [
    xml_term C\<langle>t\<rangle>,
    xml_pos (hole_pos C),
    (case l of Forbidden_Patterns.H \<Rightarrow> XML ''here'' [] [] 
             | Forbidden_Patterns.A \<Rightarrow> XML ''above'' [] [] 
             | Forbidden_Patterns.B \<Rightarrow> XML ''below'' [] [] 
             | Forbidden_Patterns.R \<Rightarrow> XML ''right'' [] [] )
   ]"

definition xml_repl_map :: "(('f :: show,('l :: show) list)lab \<times> nat) \<times> (nat list) \<Rightarrow> xml" where
  "xml_repl_map \<equiv> \<lambda> ((f,a),l). XML ''replacementMapEntry'' [] (
    xml_lab f #
    XML ''arity'' [] [XML_text (show a)] # 
    map xml_single_pos l
    )"

fun xml_strategy :: "(('f :: show,('l :: show) list)lab,'v :: show)inn_fp_strategy \<Rightarrow> xml list"
  where "xml_strategy (Inl No_Strategy) = []"
     |  "xml_strategy (Inl Innermost) = [XML ''strategy'' [] [XML ''innermost'' [] []]]"
     |  "xml_strategy (Inr Outermost) = [XML ''strategy'' [] [XML ''outermost'' [] []]]"
     |  "xml_strategy (Inr (Forbidden_Patterns p)) = [XML ''strategy'' [] [XML ''forbiddenPatterns'' [] (map xml_forbidden_pattern p)]]"
     |  "xml_strategy (Inr (Context_Sensitive \<mu>)) = [XML ''strategy'' [] [XML ''contextSensitive'' [] (map xml_repl_map \<mu>)]]"
     |  "xml_strategy (Inl (Innermost_Q Q)) = [XML ''strategy'' [] 
                                   [XML ''innermostLhss'' [] (map xml_term Q)]]"

primrec xml_trs_input :: "(('f :: show,('l :: show) list)lab,'v :: show)inn_fp_strategy \<Rightarrow> (('f,'l list)lab,'v)rules
  \<Rightarrow> (('f,'l list)lab,'v)rules option \<Rightarrow> xml"
  where "xml_trs_input strat R None =  XML ''trsInput'' [] ([xml_rules ''trs'' R] @ xml_strategy strat)"
  |     "xml_trs_input strat R (Some S) = XML ''trsInput'' [] (xml_rules ''trs'' R # xml_strategy strat @ [xml_rules ''relativeRules'' S])"

definition xml_state :: "string \<Rightarrow> xml" where
  "xml_state s = XML ''state'' [] [XML_text s]"

fun xml_eps_transition :: "string \<times> string \<Rightarrow> xml" where
  "xml_eps_transition (s,t) = XML ''transition'' [] [
    XML ''lhs'' [] [xml_state s],
    XML ''rhs'' [] [xml_state t]]"

fun xml_transition :: "(string,('f :: show,('l :: show) list)lab)ta_rule \<Rightarrow> xml" where
  "xml_transition (TA_rule f qs q) = XML ''transition'' [] [
    XML ''lhs'' [] (xml_lab f # map xml_state qs),
    XML ''rhs'' [] [xml_state q]]"

fun xml_ta :: "(string,('f :: show,('l :: show) list)lab)tree_automaton \<Rightarrow> xml" where
  "xml_ta (Tree_Automaton fin tran eps) =  XML ''treeAutomaton'' [] [
    XML ''finalStates'' [] (map xml_state fin),
    XML ''transitions'' [] (map xml_transition tran @ map xml_eps_transition eps)]"

definition xml_ta_input :: "(string,('f :: show,('l :: show) list)lab)tree_automaton \<Rightarrow> (('f,'l list)lab,'v :: show)rules
  \<Rightarrow> xml"
  where "xml_ta_input ta R =  XML ''treeAutomatonProblem'' [] ([xml_ta ta, xml_rules ''trs'' R])"

definition xml_ctrs_input :: "(('f :: show,('l :: show) list)lab,'v :: show)crules \<Rightarrow> xml"
  where "xml_ctrs_input ctrs =  XML ''ctrsInput'' [] [XML ''conditionalRules'' [] (map xml_crule ctrs)]"

primrec xml_cert_problem :: "('f :: show, ('l :: show) list, 'v :: show) cert_problem \<Rightarrow> xml" where
  "xml_cert_problem (TRS_Termination_Proof nfs Q R S_o _) = xml_trs_input (Inl Q) R S_o"
| "xml_cert_problem (TRS_Nontermination_Proof nfs Q R _) = xml_trs_input (Inl Q) R None"
| "xml_cert_problem (Outermost_Nontermination_Proof R _) = xml_trs_input (Inr Outermost) R None"
| "xml_cert_problem (Outermost_Termination_Proof R _) = xml_trs_input (Inr Outermost) R None"
| "xml_cert_problem (FP_Nontermination_Proof P R _) = xml_trs_input (Inr (Forbidden_Patterns P)) R None"
| "xml_cert_problem (FP_Termination_Proof P R _) = xml_trs_input (Inr (Forbidden_Patterns P)) R None"
| "xml_cert_problem (CS_Nontermination_Proof \<mu> R _) = xml_trs_input (Inr (Context_Sensitive \<mu>)) R None"
| "xml_cert_problem (CS_Termination_Proof \<mu> R _) = xml_trs_input (Inr (Context_Sensitive \<mu>)) R None"
| "xml_cert_problem (Relative_TRS_Nontermination_Proof nfs Q R S _) = xml_trs_input (Inl Q) R (Some S)"
| "xml_cert_problem (DP_Termination_Proof nfs m P Pw Q R Rw _) =
    XML ''dpInput'' [] (xml_rules ''trs'' Rw # xml_rules ''dps'' P # xml_strategy (Inl Q) @ [XML ''minimal'' [] [XML_text (if m then ''true'' else ''false'')]])"
| "xml_cert_problem (DP_Nontermination_Proof nfs m P Q R _) =
    XML ''dpInput'' [] (xml_rules ''trs'' R # xml_rules ''dps'' P # xml_strategy (Inl Q) @ [XML ''minimal'' [] [XML_text (if m then ''true'' else ''false'')]])"
| "xml_cert_problem (TRS_Confluence_Proof nfs R _) = xml_trs_input (Inl No_Strategy) R None"
| "xml_cert_problem (TRS_Non_Confluence_Proof nfs R _) = xml_trs_input (Inl No_Strategy) R None"
| "xml_cert_problem (Completion_Proof E R _) =
    XML ''completionInput'' [] [xml_rules ''equations'' E, xml_rules ''trs'' R]"
| "xml_cert_problem (Equational_Proof E eq _) =
    XML ''equationalReasoningInput'' [] [xml_rules ''equations'' E, XML ''equation'' [] [xml_term (fst eq), xml_term (snd eq)]]"
| "xml_cert_problem (Equational_Disproof E eq _) =
    XML ''equationalReasoningInput'' [] [xml_rules ''equations'' E, XML ''equation'' [] [xml_term (fst eq), xml_term (snd eq)]]"
| "xml_cert_problem (Complexity_Proof Q R S cm cc _) =
    XML ''complexityInput'' [] [xml_trs_input (Inl Q) R S, xml_complexity_measure cm, xml_complexity_class cc]"
| "xml_cert_problem (Quasi_Reductive_Proof ctrs _) = xml_ctrs_input ctrs"
| "xml_cert_problem (Conditional_CR_Proof ctrs _) = xml_ctrs_input ctrs"
| "xml_cert_problem (Tree_Automata_Closed_Proof ta r _) = xml_ta_input ta r"
| "xml_cert_problem (Unknown_Proof u _) = XML ''unknownInput'' [] [XML_text u]"
| "xml_cert_problem (Unknown_Disproof u _) = XML ''unknownInput'' [] [XML_text u]"

end
